"""Provenance tracking utilities for compiler objects."""

__all__ = [
    "Note",
    "NoteKind",
    "Position",
    "Provenance",
    "Span",
]

from dataclasses import dataclass
from pathlib import Path
from typing import TypedDict, TypeGuard

from fhy_core.serialization import (
    DeserializationDictStructureError,
    DeserializationValueError,
    Serializable,
    SerializedDict,
    is_serialized_dict,
    register_serializable,
)
from fhy_core.trait.equality import EqualMixin
from fhy_core.trait.frozen import FrozenMixin
from fhy_core.utils import StrEnum


class _PositionData(TypedDict):
    line: int
    column: int


def _is_valid_position_data(data: SerializedDict) -> TypeGuard[_PositionData]:
    return (
        "line" in data
        and isinstance(data["line"], int)
        and "column" in data
        and isinstance(data["column"], int)
    )


@register_serializable(type_id="position")
@dataclass(frozen=True, slots=True, order=True)
class Position(Serializable, FrozenMixin, EqualMixin):
    """Line/column position."""

    line: int
    column: int

    @property
    def supports_ordering(self) -> bool:
        return True

    @property
    def supports_partial_ordering(self) -> bool:
        return True

    def __post_init__(self) -> None:
        if self.line < 1:
            raise ValueError(f'"line" must be >= 1, got {self.line}')
        if self.column < 1:
            raise ValueError(f'"column" must be >= 1, got {self.column}')

    def serialize_to_dict(self) -> SerializedDict:
        return {
            "line": self.line,
            "column": self.column,
        }

    @classmethod
    def deserialize_from_dict(cls, data: SerializedDict) -> "Position":
        if not _is_valid_position_data(data):
            raise DeserializationDictStructureError(
                cls, _PositionData.__annotations__, data
            )
        try:
            return cls(data["line"], data["column"])
        except ValueError as exc:
            raise DeserializationValueError(f"Invalid position values: {exc}") from exc

    def __str__(self) -> str:
        return f"{self.line}:{self.column}"


class _SpanData(TypedDict):
    file_path: str
    start_offset: int | None
    end_offset: int | None
    start_position: _PositionData | None
    end_position: _PositionData | None


def _is_valid_span_data(data: SerializedDict) -> TypeGuard[_SpanData]:
    if not is_serialized_dict(data):
        return False
    if "file_path" not in data or not isinstance(data["file_path"], str):
        return False
    if "start_offset" not in data or not (
        isinstance(data["start_offset"], int) or data["start_offset"] is None
    ):
        return False
    if "end_offset" not in data or not (
        isinstance(data["end_offset"], int) or data["end_offset"] is None
    ):
        return False
    start_position_data = data["start_position"] if "start_position" in data else None
    if "start_position" not in data or not (
        (
            is_serialized_dict(start_position_data)
            and _is_valid_position_data(start_position_data)
        )
        or start_position_data is None
    ):
        return False
    end_position_data = data["end_position"] if "end_position" in data else None
    if "end_position" not in data or not (
        (
            is_serialized_dict(end_position_data)
            and _is_valid_position_data(end_position_data)
        )
        or end_position_data is None
    ):
        return False
    return True


@register_serializable(type_id="span")
@dataclass(frozen=True, slots=True)
class Span(Serializable, FrozenMixin, EqualMixin):
    """A region in a source file."""

    file_path: Path
    start_offset: int | None = None
    end_offset: int | None = None
    start_position: Position | None = None
    end_position: Position | None = None

    def __post_init__(self) -> None:
        if self.start_offset is not None and self.start_offset < 0:
            raise ValueError(f'"start_offset" must be >= 0, got {self.start_offset}')
        if self.end_offset is not None and self.end_offset < 0:
            raise ValueError(f'"end_offset" must be >= 0, got {self.end_offset}')
        if (
            self.start_offset is not None
            and self.end_offset is not None
            and self.end_offset < self.start_offset
        ):
            raise ValueError(
                '"end_offset" must be >= "start_offset", got '
                f"{self.end_offset} < {self.start_offset}"
            )
        if (
            self.start_position is not None
            and self.end_position is not None
            and self.end_position < self.start_position
        ):
            raise ValueError(
                '"end_position" must be >= "start_position", got '
                f"{self.end_position} < {self.start_position}"
            )

    def is_unknown(self) -> bool:
        return (
            self.start_offset is None
            and self.end_offset is None
            and self.start_position is None
            and self.end_position is None
        )

    def serialize_to_dict(self) -> SerializedDict:
        return {
            "file_path": str(self.file_path),
            "start_offset": self.start_offset,
            "end_offset": self.end_offset,
            "start_position": self.start_position.serialize_to_dict()
            if self.start_position
            else None,
            "end_position": self.end_position.serialize_to_dict()
            if self.end_position
            else None,
        }

    @classmethod
    def deserialize_from_dict(cls, data: SerializedDict) -> "Span":
        if not _is_valid_span_data(data):
            raise DeserializationDictStructureError(
                cls, _SpanData.__annotations__, data
            )

        start_position_data = data["start_position"]
        end_position_data = data["end_position"]

        try:
            return cls(
                Path(data["file_path"]),
                start_offset=data["start_offset"],
                end_offset=data["end_offset"],
                start_position=(
                    Position.deserialize_from_dict(start_position_data)
                    if is_serialized_dict(start_position_data)
                    else None
                ),
                end_position=(
                    Position.deserialize_from_dict(end_position_data)
                    if is_serialized_dict(end_position_data)
                    else None
                ),
            )
        except ValueError as exc:
            raise DeserializationValueError(f"Invalid span values: {exc}") from exc

    def __str__(self) -> str:
        file_path = str(self.file_path)
        if self.start_position and self.end_position:
            return f"{file_path}:{self.start_position}-{self.end_position}"
        if self.start_offset is not None and self.end_offset is not None:
            return f"{file_path}@{self.start_offset}-{self.end_offset}"
        return file_path


class NoteKind(StrEnum):
    """Structured note kinds so tooling can filter/group notes."""

    OTHER = "other"


class _NoteData(TypedDict):
    message: str
    kind: str


def _is_valid_note_data(data: SerializedDict) -> TypeGuard[_NoteData]:
    return (
        "message" in data
        and isinstance(data["message"], str)
        and "kind" in data
        and isinstance(data["kind"], str)
    )


@register_serializable(type_id="provenance_note")
@dataclass(frozen=True, slots=True)
class Note(Serializable, FrozenMixin, EqualMixin):
    """A provenance breadcrumb."""

    message: str
    kind: NoteKind = NoteKind.OTHER

    def serialize_to_dict(self) -> SerializedDict:
        return {"message": self.message, "kind": self.kind.value}

    @classmethod
    def deserialize_from_dict(cls, data: SerializedDict) -> "Note":
        if not _is_valid_note_data(data):
            raise DeserializationDictStructureError(
                cls, _NoteData.__annotations__, data
            )
        try:
            return cls(data["message"], kind=NoteKind(data["kind"]))
        except ValueError as exc:
            raise DeserializationValueError(f"Invalid note values: {exc}") from exc

    def __str__(self) -> str:
        return f"{self.kind}: {self.message}"


class _ProvenanceData(TypedDict):
    span: _SpanData | None
    origins: list[SerializedDict]
    notes: list[SerializedDict]
    source_ids: list[str]


def _is_valid_provenance_data(data: SerializedDict) -> TypeGuard[_ProvenanceData]:
    if not is_serialized_dict(data):
        return False
    if "span" not in data or not (
        is_serialized_dict(data["span"])
        and _is_valid_span_data(data["span"])
        or data["span"] is None
    ):
        return False
    if "origins" not in data or not isinstance(data["origins"], list):
        return False
    for origin in data["origins"]:
        if not is_serialized_dict(origin) or not _is_valid_span_data(origin):
            return False
    if "notes" not in data or not isinstance(data["notes"], list):
        return False
    for note in data["notes"]:
        if not is_serialized_dict(note) or not _is_valid_note_data(note):
            return False
    if "source_ids" not in data or not isinstance(data["source_ids"], list):
        return False
    for source_id in data["source_ids"]:
        if not isinstance(source_id, str):
            return False
    return True


@register_serializable(type_id="provenance")
@dataclass(frozen=True, slots=True)
class Provenance(Serializable, FrozenMixin, EqualMixin):
    """Immutable provenance for compiler objects."""

    span: Span | None = None
    origins: tuple[Span, ...] = ()
    notes: tuple[Note, ...] = ()
    source_ids: tuple[str, ...] = ()

    @staticmethod
    def unknown() -> "Provenance":
        """Return a provenance with no information."""
        return Provenance()

    def with_span(self, span: Span) -> "Provenance":
        """Return a new provenance with the span, preserving other metadata."""
        return Provenance(
            span=span,
            origins=self.origins,
            notes=self.notes,
            source_ids=self.source_ids,
        )

    def add_origin(self, origin: Span) -> "Provenance":
        """Return a new provenance with the origin, preserving other metadata."""
        return Provenance(
            span=self.span,
            origins=self.origins + (origin,),
            notes=self.notes,
            source_ids=self.source_ids,
        )

    def add_note(self, note: Note) -> "Provenance":
        """Return a new provenance with the note, preserving other metadata."""
        return Provenance(
            span=self.span,
            origins=self.origins,
            notes=self.notes + (note,),
            source_ids=self.source_ids,
        )

    def add_source_id(self, source_id: str) -> "Provenance":
        """Return a new provenance with the source ID, preserving other metadata."""
        return Provenance(
            span=self.span,
            origins=self.origins,
            notes=self.notes,
            source_ids=self.source_ids + (source_id,),
        )

    def merge(self, *others: "Provenance") -> "Provenance":
        """Merge multiple provenances while preserving insertion order."""
        span = self.span
        origins = list(self.origins)
        notes = list(self.notes)
        source_ids = list(self.source_ids)

        for other in others:
            if span is None and other.span is not None:
                span = other.span
            origins.extend(other.origins)
            notes.extend(other.notes)
            source_ids.extend(other.source_ids)

        return Provenance(
            span=span,
            origins=tuple(origins),
            notes=tuple(notes),
            source_ids=tuple(source_ids),
        )

    def serialize_to_dict(self) -> SerializedDict:
        return {
            "span": self.span.serialize_to_dict() if self.span else None,
            "origins": [origin.serialize_to_dict() for origin in self.origins],
            "notes": [note.serialize_to_dict() for note in self.notes],
            "source_ids": list(self.source_ids),
        }

    @classmethod
    def deserialize_from_dict(cls, data: SerializedDict) -> "Provenance":
        if not _is_valid_provenance_data(data):
            raise DeserializationDictStructureError(
                cls, _ProvenanceData.__annotations__, data
            )

        span_data = data["span"]
        return cls(
            span=Span.deserialize_from_dict(span_data)
            if is_serialized_dict(span_data)
            else None,
            origins=tuple(
                Span.deserialize_from_dict(origin) for origin in data["origins"]
            ),
            notes=tuple(Note.deserialize_from_dict(note) for note in data["notes"]),
            source_ids=tuple(data["source_ids"]),
        )



"""Bound numerical parameters."""

__all__ = ["BoundIntParam", "BoundNatParam"]

from typing import Any, Iterable, Optional, Tuple, TypeGuard, cast

from fhy_core.constraint import Constraint, EquationConstraint
from fhy_core.expression import (
    BinaryExpression,
    BinaryOperation,
    Expression,
    IdentifierExpression,
    LiteralExpression,
)
from fhy_core.identifier import Identifier
from fhy_core.serialization import (
    DeserializationDictStructureError,
    DeserializationValueError,
    SerializedDict,
    register_serializable,
)
from fhy_core.utils import Self

from .core import (
    IntParam,
    ParamData,
    finalize_param_construction_from_data,
    is_valid_param_data,
)
from .fundamental import (
    NatParam,
    is_the_basic_nat_param_constraint,
    is_zero_included_constraint_exists,
    is_zero_not_included_constraint_exists,
)


def _is_valid_bound_expression(expression: Expression) -> bool:
    if not isinstance(expression, BinaryExpression):
        return False
    if expression.operation not in (
        BinaryOperation.GREATER_EQUAL,
        BinaryOperation.GREATER,
        BinaryOperation.LESS_EQUAL,
        BinaryOperation.LESS,
    ):
        return False
    if not (
        (
            isinstance(expression.left, IdentifierExpression)
            or isinstance(expression.right, IdentifierExpression)
        )
        and (
            isinstance(expression.left, LiteralExpression)
            or isinstance(expression.right, LiteralExpression)
        )
    ):
        return False
    literal_expression: LiteralExpression
    if isinstance(expression.left, LiteralExpression):
        literal_expression = expression.left
    elif isinstance(expression.right, LiteralExpression):
        literal_expression = expression.right
    else:
        raise RuntimeError(
            "Somehow failed to find LiteralExpression in bound expression."
        )
    return isinstance(literal_expression.value, int)


def _invert_binary_comparison_operation(op: BinaryOperation) -> BinaryOperation:
    if op == BinaryOperation.GREATER:
        return BinaryOperation.LESS
    elif op == BinaryOperation.GREATER_EQUAL:
        return BinaryOperation.LESS_EQUAL
    elif op == BinaryOperation.LESS:
        return BinaryOperation.GREATER
    elif op == BinaryOperation.LESS_EQUAL:
        return BinaryOperation.GREATER_EQUAL
    else:
        raise ValueError(f"Cannot invert non-comparison binary operation: {op}")


def _get_bound_from_expression(
    literal_expression: LiteralExpression, op: BinaryOperation
) -> Tuple[bool, int, bool]:
    value = literal_expression.value
    if not isinstance(value, int):
        raise RuntimeError(
            "Somehow failed to find integer LiteralExpression in bound expression."
        )
    return (
        True
        if op in (BinaryOperation.GREATER, BinaryOperation.GREATER_EQUAL)
        else False,
        value,
        True
        if op in (BinaryOperation.GREATER_EQUAL, BinaryOperation.LESS_EQUAL)
        else False,
    )


class _BoundParamData(ParamData):
    prefer_inclusive: bool


def _is_valid_bound_param_data(data: SerializedDict) -> TypeGuard[_BoundParamData]:
    return (
        "prefer_inclusive" in data
        and isinstance(data["prefer_inclusive"], bool)
        and is_valid_param_data(data)
    )


@register_serializable(type_id="bound_int_param")
class BoundIntParam(IntParam):
    """An interval integer parameter that supports basic arithmetic.

    Results can be represented as inclusive or exclusive via prefer_inclusive.
    """

    _prefer_inclusive: bool

    def __init__(
        self,
        *,
        name: Identifier | None = None,
        prefer_inclusive: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(name=name, **kwargs)
        object.__setattr__(self, "_prefer_inclusive", prefer_inclusive)

    def _validate_constraint(self, constraint: Constraint) -> None:
        super()._validate_constraint(constraint)
        if not isinstance(constraint, EquationConstraint):
            raise TypeError(
                "BoundIntParam only supports EquationConstraint constraints."
            )
        if not _is_valid_bound_expression(constraint.convert_to_expression()):
            raise ValueError(
                "BoundIntParam only supports bound expressions of the form "
                '"x >= k", "x > k", "x <= k", or "x < k" where k is an integer.'
            )

    def _iter_bounds(self) -> Iterable[Tuple[bool, int, bool]]:
        """Yield (is_lower, bound_value, inclusive) for constraints.

        Fail if any constraint is not a valid bound expression on the same
        variable.
        """
        for constraint in self._constraints:
            if not isinstance(constraint, EquationConstraint):
                raise RuntimeError(
                    "BoundIntParam somehow has non-EquationConstraint constraint: "
                    f"{type(constraint)}"
                )
            if not _is_valid_bound_expression(constraint.convert_to_expression()):
                raise RuntimeError(
                    "BoundIntParam somehow has non-bound expression constraint: "
                    f"{repr(constraint)}"
                )

            expression = constraint.convert_to_expression()
            identifier_expression: IdentifierExpression | None = None

            def set_or_check_identifier_expr(e: IdentifierExpression) -> None:
                nonlocal identifier_expression
                if identifier_expression is None:
                    identifier_expression = e
                elif identifier_expression != e:
                    raise RuntimeError(
                        "BoundIntParam somehow has bound constraints on "
                        "different variables."
                    )

            if not isinstance(expression, BinaryExpression):
                raise RuntimeError(
                    "Somehow bound constraint expression is not a BinaryExpression."
                )
            if isinstance(expression.left, IdentifierExpression):
                if not isinstance(expression.right, LiteralExpression):
                    raise RuntimeError(
                        "Somehow bound expression is not in the expected form."
                    )
                set_or_check_identifier_expr(expression.left)
                yield _get_bound_from_expression(expression.right, expression.operation)
            elif isinstance(expression.right, IdentifierExpression):
                if not isinstance(expression.left, LiteralExpression):
                    raise RuntimeError(
                        "Somehow bound expression is not in the expected form."
                    )
                set_or_check_identifier_expr(expression.right)
                yield _get_bound_from_expression(
                    expression.left,
                    _invert_binary_comparison_operation(expression.operation),
                )
            else:
                raise RuntimeError(
                    "Somehow bound expression is not in the expected form."
                )

    def _get_effective_min_max(self) -> Tuple[int | None, int | None]:
        """Return (min_int, max_int) represented by constraints.

        Semantics:
          x > k  => min_int >= k+1
          x >= k => min_int >= k
          x < k  => max_int <= k-1
          x <= k => max_int <= k
        None means unbounded.
        """
        min_int: int | None = None
        max_int: int | None = None

        for is_lower, bound, inclusive in self._iter_bounds():
            if is_lower:
                eff = bound if inclusive else bound + 1
                min_int = eff if min_int is None else max(min_int, eff)
            else:
                eff = bound if inclusive else bound - 1
                max_int = eff if max_int is None else min(max_int, eff)

        if min_int is not None and max_int is not None and min_int > max_int:
            raise ValueError(
                "Empty integer interval represented by constraints for "
                f"{self.variable}."
            )

        return min_int, max_int

    @staticmethod
    def _create_param_from_min_max(
        param: "BoundIntParam",
        min_int: Optional[int],
        max_int: Optional[int],
    ) -> "BoundIntParam":
        """Create a new BoundIntParam from min/max integers.

        Emits either inclusive or exclusive style depending on param.prefer_inclusive.
        """
        out = BoundIntParam(
            name=param.variable, prefer_inclusive=param._prefer_inclusive
        )
        if min_int is not None:
            if out._prefer_inclusive:
                out = out.add_lower_bound_constraint(min_int, is_inclusive=True)
            else:
                out = out.add_lower_bound_constraint(min_int - 1, is_inclusive=False)

        if max_int is not None:
            if out._prefer_inclusive:
                out = out.add_upper_bound_constraint(max_int, is_inclusive=True)
            else:
                out = out.add_upper_bound_constraint(max_int + 1, is_inclusive=False)
        return out

    @classmethod
    def between(
        cls: type[Self],
        lower_bound: int,
        upper_bound: int,
        *,
        name: Identifier | None = None,
        is_lower_inclusive: bool = True,
        is_upper_inclusive: bool = True,
        prefer_inclusive: bool = True,
    ) -> "BoundIntParam":
        p = cls(name=name, prefer_inclusive=prefer_inclusive)
        p = p.add_lower_bound_constraint(lower_bound, is_inclusive=is_lower_inclusive)
        p = p.add_upper_bound_constraint(upper_bound, is_inclusive=is_upper_inclusive)
        p._get_effective_min_max()
        return p

    @classmethod
    def with_lower_bound(
        cls: type[Self],
        lower_bound: int,
        *,
        name: Identifier | None = None,
        is_inclusive: bool = True,
        prefer_inclusive: bool = True,
    ) -> "BoundIntParam":
        p = cls(name=name, prefer_inclusive=prefer_inclusive)
        p = p.add_lower_bound_constraint(lower_bound, is_inclusive=is_inclusive)
        return p

    @classmethod
    def with_upper_bound(
        cls: type[Self],
        upper_bound: int,
        *,
        name: Identifier | None = None,
        is_inclusive: bool = True,
        prefer_inclusive: bool = True,
    ) -> "BoundIntParam":
        p = cls(name=name, prefer_inclusive=prefer_inclusive)
        p = p.add_upper_bound_constraint(upper_bound, is_inclusive=is_inclusive)
        return p

    @classmethod
    def exactly(
        cls: type[Self],
        value: int,
        *,
        name: Identifier | None = None,
        prefer_inclusive: bool = True,
    ) -> "BoundIntParam":
        p = cls(name=name, prefer_inclusive=prefer_inclusive)
        p = p.add_lower_bound_constraint(value, is_inclusive=True)
        p = p.add_upper_bound_constraint(value, is_inclusive=True)
        return p

    def serialize_data_to_dict(self) -> SerializedDict:
        super_dict = super().serialize_data_to_dict()
        super_dict["prefer_inclusive"] = self._prefer_inclusive
        return super_dict

    def _clone(self) -> Self:
        init_kwargs: dict[str, Any] = {
            "name": self._variable,
            "prefer_inclusive": self._prefer_inclusive,
        }
        if hasattr(self, "_is_zero_included"):
            init_kwargs["is_zero_included"] = getattr(self, "_is_zero_included")
        new_param = self.__class__(**init_kwargs)
        object.__setattr__(new_param, "_constraints", self._constraints)
        return new_param

    @classmethod
    def deserialize_data_from_dict(cls, data: SerializedDict) -> "BoundIntParam":
        if not _is_valid_bound_param_data(data):
            raise DeserializationDictStructureError(
                cls, _BoundParamData.__annotations__, data
            )
        param = BoundIntParam(
            name=Identifier.deserialize_from_dict(data["variable"]),
            prefer_inclusive=data["prefer_inclusive"],
        )
        return cast(
            BoundIntParam,
            finalize_param_construction_from_data(param, data),
        )

    def _coerce_other(self, other: Any) -> "BoundIntParam":
        if isinstance(other, int):
            return BoundIntParam.exactly(other, prefer_inclusive=self._prefer_inclusive)
        elif isinstance(other, BoundIntParam):
            return other
        elif isinstance(other, IntParam):
            for other_constraint in other._constraints:
                if not _is_valid_bound_expression(
                    other_constraint.convert_to_expression()
                ):
                    raise TypeError(
                        "Cannot coerce IntParam with non-bound constraints "
                        "to BoundIntParam."
                    )
            wrapper_param = BoundIntParam(
                name=other.variable, prefer_inclusive=self._prefer_inclusive
            )
            object.__setattr__(wrapper_param, "_constraints", other._constraints)
            return wrapper_param
        raise TypeError(f"Unsupported operand type: {type(other)}")

    def __add__(self, other: Any) -> "BoundIntParam":
        coerced_other = self._coerce_other(other)

        self_min, self_max = self._get_effective_min_max()
        other_min, other_max = coerced_other._get_effective_min_max()
        new_min = (
            None if (self_min is None or other_min is None) else self_min + other_min
        )
        new_max = (
            None if (self_max is None or other_max is None) else self_max + other_max
        )
        return self._create_param_from_min_max(self, new_min, new_max)

    def __radd__(self, other: Any) -> "BoundIntParam":
        return self.__add__(other)

    def __sub__(self, other: Any) -> "BoundIntParam":
        coerced_other = self._coerce_other(other)

        self_min, self_max = self._get_effective_min_max()
        other_min, other_max = coerced_other._get_effective_min_max()
        new_min = (
            None if (self_min is None or other_max is None) else self_min - other_max
        )
        new_max = (
            None if (self_max is None or other_min is None) else self_max - other_min
        )
        return self._create_param_from_min_max(self, new_min, new_max)

    def __rsub__(self, other: Any) -> "BoundIntParam":
        return self._coerce_other(other).__sub__(self)

    def __neg__(self) -> "BoundIntParam":
        self_min, self_max = self._get_effective_min_max()
        new_min = None if self_max is None else -self_max
        new_max = None if self_min is None else -self_min
        return self._create_param_from_min_max(self, new_min, new_max)


@register_serializable(type_id="bound_nat_param")
class BoundNatParam(BoundIntParam, NatParam):
    """A bounded natural number parameter (i.e., non-negative integers)."""

    def __init__(
        self,
        *,
        name: Identifier | None = None,
        is_zero_included: bool = True,
        prefer_inclusive: bool = True,
    ) -> None:
        super().__init__(
            name=name,
            is_zero_included=is_zero_included,
            prefer_inclusive=prefer_inclusive,
        )

    @classmethod
    def deserialize_data_from_dict(cls, data: SerializedDict) -> "BoundNatParam":
        if not _is_valid_bound_param_data(data):
            raise DeserializationDictStructureError(
                cls, _BoundParamData.__annotations__, data
            )

        constraints = [Constraint.deserialize_from_dict(c) for c in data["constraints"]]
        variable = Identifier.deserialize_from_dict(data["variable"])
        prefer_inclusive = data["prefer_inclusive"]

        if is_zero_included_constraint_exists(constraints, variable):
            is_zero_included = True
        elif is_zero_not_included_constraint_exists(constraints, variable):
            is_zero_included = False
        else:
            raise DeserializationValueError(
                cls,
                "constraints",
                (
                    "a constraint bounding the variable to be greater than or "
                    "equal to zero or greater than zero"
                ),
                data["constraints"],
            )

        param = BoundNatParam(
            name=variable,
            is_zero_included=is_zero_included,
            prefer_inclusive=prefer_inclusive,
        )
        return cast(
            BoundNatParam,
            finalize_param_construction_from_data(
                param,
                data,
                constraint_filter_function=lambda c: (
                    not is_the_basic_nat_param_constraint(c, variable, is_zero_included)
                ),
            ),
        )



"""`Frozen` trait and mixin."""

__all__ = [
    "Frozen",
    "FrozenMixin",
    "FrozenMutationError",
    "FrozenValidationError",
]

from abc import ABC
from dataclasses import FrozenInstanceError, is_dataclass
from types import MappingProxyType
from typing import Any, Protocol, runtime_checkable

from frozendict import frozendict

from fhy_core.error import register_error

_FROZEN_FLAG = "_fhy_core_is_frozen"
_IMMUTABLE_ATOMS = (str, bytes, int, float, complex, bool, type(None))


@register_error
class FrozenMutationError(RuntimeError):
    """Raised when mutating a frozen object."""


@register_error
class FrozenValidationError(RuntimeError):
    """Raised when runtime frozen verification fails."""


@runtime_checkable
class Frozen(Protocol):
    """Protocol for objects that support runtime freezing."""

    @property
    def is_frozen(self) -> bool: ...

    def freeze(self, *, deep: bool = False) -> None: ...

    def assert_frozen(self, *, deep: bool = False, strict: bool = False) -> None: ...

    def assert_write_protected(self) -> None: ...


class FrozenMixin(ABC):
    """Mixin that enforces runtime immutability after `freeze()`."""

    @property
    def is_frozen(self) -> bool:
        return self._is_native_frozen_dataclass() or self._is_marked_frozen()

    def freeze(self, *, deep: bool = False) -> None:
        """Freeze this object and optionally deep-freeze its instance state.

        Args:
            deep: Whether to recursively freeze nested objects in the instance state.

        """
        if self._is_native_frozen_dataclass():
            if not deep:
                return
            seen: set[int] = {id(self)}
            for attribute_name, attribute_value in self._iter_instance_state_items():
                if attribute_name == _FROZEN_FLAG:
                    continue
                frozen_value = self._freeze_value(attribute_value, seen)
                object.__setattr__(self, attribute_name, frozen_value)
            return

        if self._is_marked_frozen():
            return
        object.__setattr__(self, _FROZEN_FLAG, True)
        if deep:
            seen_local: set[int] = {id(self)}
            for attribute_name, attribute_value in self._iter_instance_state_items():
                if attribute_name == _FROZEN_FLAG:
                    continue
                frozen_value = self._freeze_value(attribute_value, seen_local)
                object.__setattr__(self, attribute_name, frozen_value)

    def assert_frozen(self, *, deep: bool = False, strict: bool = False) -> None:
        """Assert that this object is frozen.

        Args:
            deep: Whether to verify deep immutability for the instance state.
            strict: Whether unknown nested object types are treated as mutable.

        Raises:
            FrozenValidationError: If frozen constraints are violated.

        """
        if not self.is_frozen:
            raise FrozenValidationError(f"{type(self).__name__} is not frozen.")
        self.assert_write_protected()
        if deep and not self._is_deeply_immutable(self, set(), strict=strict):
            raise FrozenValidationError(
                f'"{type(self).__name__}" failed deep immutability validation.'
            )

    def assert_write_protected(self) -> None:
        """Assert that normal `setattr` attempts are blocked for this object."""
        if not self.is_frozen:
            raise FrozenValidationError(
                f"{type(self).__name__} is not frozen, so write protection is inactive."
            )
        candidate_name: str | None = None
        for attribute_name, _ in self._iter_instance_state_items():
            if attribute_name == _FROZEN_FLAG:
                continue
            candidate_name = attribute_name
            break
        if candidate_name is None:
            return
        try:
            setattr(self, candidate_name, getattr(self, candidate_name))
        except FrozenMutationError:
            return
        except FrozenInstanceError:
            return
        except Exception as exc:
            raise FrozenValidationError(
                f'"{type(self).__name__}" write-protection probe failed '
                f"unexpectedly: {exc}"
            ) from exc
        raise FrozenValidationError(
            f'"{type(self).__name__}" accepted mutation while frozen.'
        )

    def __setattr__(self, name: str, value: Any) -> None:
        if name in {_FROZEN_FLAG, "__orig_class__"}:
            object.__setattr__(self, name, value)
            return
        if self._is_marked_frozen():
            raise FrozenMutationError(
                f'Cannot modify "{name}" on frozen {type(self).__name__}.'
            )
        object.__setattr__(self, name, value)

    def __delattr__(self, name: str) -> None:
        if name != _FROZEN_FLAG and self._is_marked_frozen():
            raise FrozenMutationError(
                f'Cannot delete "{name}" on frozen {type(self).__name__}.'
            )
        object.__delattr__(self, name)

    @staticmethod
    def _freeze_value(value: Any, seen: set[int]) -> Any:
        value_id = id(value)
        if value_id in seen:
            return value
        seen.add(value_id)

        if isinstance(value, list):
            return tuple(FrozenMixin._freeze_value(item, seen) for item in value)
        elif isinstance(value, tuple):
            return tuple(FrozenMixin._freeze_value(item, seen) for item in value)
        elif isinstance(value, set):
            return frozenset(FrozenMixin._freeze_value(item, seen) for item in value)
        elif isinstance(value, frozenset):
            return frozenset(FrozenMixin._freeze_value(item, seen) for item in value)
        elif isinstance(value, dict):
            frozen_dict = {
                FrozenMixin._freeze_value(key, seen): FrozenMixin._freeze_value(
                    item_value, seen
                )
                for key, item_value in value.items()
            }
            return frozendict(frozen_dict)
        elif isinstance(value, MappingProxyType):
            frozen_dict = {
                FrozenMixin._freeze_value(key, seen): FrozenMixin._freeze_value(
                    item_value, seen
                )
                for key, item_value in value.items()
            }
            return frozendict(frozen_dict)
        elif isinstance(value, bytearray):
            return bytes(value)
        elif isinstance(value, FrozenMixin):
            if value.is_frozen:
                return value
            value.freeze(deep=True)
            return value
        else:
            return value

    @staticmethod
    def _is_deeply_immutable(value: Any, seen: set[int], *, strict: bool) -> bool:
        value_id = id(value)
        if value_id in seen:
            return True
        seen.add(value_id)

        if isinstance(value, _IMMUTABLE_ATOMS):
            return True
        elif isinstance(value, tuple):
            return all(
                FrozenMixin._is_deeply_immutable(item, seen, strict=strict)
                for item in value
            )
        elif isinstance(value, frozenset):
            return all(
                FrozenMixin._is_deeply_immutable(item, seen, strict=strict)
                for item in value
            )
        elif isinstance(value, (frozendict, MappingProxyType)):
            return all(
                FrozenMixin._is_deeply_immutable(key, seen, strict=strict)
                and FrozenMixin._is_deeply_immutable(item_value, seen, strict=strict)
                for key, item_value in value.items()
            )
        elif isinstance(value, FrozenMixin):
            if not value.is_frozen:
                return False
            return all(
                FrozenMixin._is_deeply_immutable(item_value, seen, strict=strict)
                for _, item_value in value._iter_instance_state_items()
                if _ != _FROZEN_FLAG
            )
        elif isinstance(value, (list, set, dict, bytearray)):
            return False
        elif strict and FrozenMixin._has_instance_state(value):
            return False
        else:
            return True

    def _iter_instance_state_items(self) -> list[tuple[str, Any]]:
        state: dict[str, Any] = {}
        instance_dict = getattr(self, "__dict__", None)
        if isinstance(instance_dict, dict):
            state.update(instance_dict)

        for klass in type(self).mro():
            slots = getattr(klass, "__slots__", ())
            if isinstance(slots, str):
                slot_names = (slots,)
            else:
                slot_names = tuple(slots)
            for slot_name in slot_names:
                if slot_name in {"__dict__", "__weakref__"}:
                    continue
                if slot_name in state:
                    continue
                try:
                    state[slot_name] = object.__getattribute__(self, slot_name)
                except AttributeError:
                    continue

        return list(state.items())

    def _is_marked_frozen(self) -> bool:
        try:
            return bool(object.__getattribute__(self, _FROZEN_FLAG))
        except AttributeError:
            return False

    def _is_native_frozen_dataclass(self) -> bool:
        params = getattr(type(self), "__dataclass_params__", None)
        return bool(is_dataclass(self) and params is not None and params.frozen)

    @staticmethod
    def _has_instance_state(value: Any) -> bool:
        if hasattr(value, "__dict__"):
            return True
        for klass in type(value).mro():
            slots = getattr(klass, "__slots__", ())
            if isinstance(slots, str):
                if slots not in {"__dict__", "__weakref__"}:
                    return True
            else:
                for slot_name in slots:
                    if slot_name not in {"__dict__", "__weakref__"}:
                        return True
        return False



"""Tests the identifier."""

from concurrent.futures import ThreadPoolExecutor
from unittest.mock import patch

import pytest
from fhy_core.identifier import Identifier
from fhy_core.serialization import (
    DeserializationDictStructureError,
    DeserializationValueError,
)
from fhy_core.trait import Equal, PartialEqual


def test_identifier_initialization():
    """Test that the identifier is initialized correctly."""
    identifier = Identifier("test_name")
    assert identifier.name_hint == "test_name"
    assert isinstance(identifier.id, int)


@patch.object(Identifier, "_next_id", 0)
def test_unique_id_generation():
    """Test that the identifier generates a unique ID."""
    id1 = Identifier("")
    id2 = Identifier("")
    id3 = Identifier("")

    assert id1.id == 0
    assert id2.id == 1
    assert id3.id == 2


def test_unique_id_generation_thread_safe():
    """Test that IDs remain unique under concurrent creation."""
    num_identifiers = 2000
    with patch.object(Identifier, "_next_id", 0):
        with ThreadPoolExecutor(max_workers=32) as executor:
            identifiers = list(
                executor.map(lambda _: Identifier(""), range(num_identifiers))
            )

    ids = [identifier.id for identifier in identifiers]
    assert len(set(ids)) == num_identifiers
    assert min(ids) == 0
    assert max(ids) == num_identifiers - 1


def test_equality():
    """Test that the identifier equality is based on the ID."""
    id1 = Identifier("name")
    id2 = Identifier("name")
    id3 = Identifier("bar")

    assert id1 == id1  # noqa: PLR0124
    assert id1 != id2
    assert id2 != id3


def test_identifier_supports_equal_traits() -> None:
    """Test `Identifier` satisfies equality trait protocols."""
    identifier = Identifier("name")
    assert isinstance(identifier, PartialEqual)
    assert isinstance(identifier, Equal)
    assert identifier.supports_partial_equality is True
    assert identifier.supports_equality is True


def test_string_representation():
    """Test that the identifier string representation is correct."""
    identifier = Identifier("test_repr")
    assert str(identifier) == "test_repr"
    assert repr(identifier) == f"test_repr::{identifier.id}"


def test_hash():
    """Test that the identifier hash is based on the ID."""
    identifier = Identifier("hash_test")
    assert hash(identifier) == hash(identifier.id)


def test_dict_serialization():
    """Test the identifier can be serialized/deserialized via a dictionary."""
    identifier = Identifier("serialization_test")
    serialized = identifier.serialize_to_dict()
    deserialized = Identifier.deserialize_from_dict(serialized)

    assert deserialized == identifier

    with pytest.raises(DeserializationDictStructureError):
        Identifier.deserialize_from_dict({"invalid": "data"})
    with pytest.raises(DeserializationDictStructureError):
        Identifier.deserialize_from_dict({"id": "not_an_int", "name_hint": "test"})
    with pytest.raises(DeserializationDictStructureError):
        Identifier.deserialize_from_dict({"id": 1, "name_hint": 123})
    with pytest.raises(DeserializationValueError):
        Identifier.deserialize_from_dict({"id": -1, "name_hint": "test"})



"""Lattice (order theory) utility."""

__all__ = ["Lattice"]

from typing import Generic, TypeVar

from .poset import PartiallyOrderedSet

T = TypeVar("T")


class Lattice(Generic[T]):
    """Lattice (order theory)."""

    _poset: PartiallyOrderedSet[T]

    def __init__(self) -> None:
        self._poset = PartiallyOrderedSet[T]()

    def __contains__(self, element: T) -> bool:
        return element in self._poset

    def add_element(self, element: T) -> None:
        """Add an element to the lattice.

        Args:
            element: The element to add.

        Raises:
            ValueError: If the element is already in the lattice.

        """
        self._poset.add_element(element)

    def add_order(self, lower: T, upper: T) -> None:
        """Add an order relation between two elements.

        Args:
            lower: The lesser element.
            upper: The greater element.

        Raises:
            ValueError: If either the lower or upper element is not in the
                lattice.
            RuntimeError: If the order relation is invalid.

        """
        self._poset.add_order(lower, upper)

    def is_lattice(self) -> bool:
        """Return True if the lattice is a valid lattice, False otherwise."""
        for x in self._poset:
            for y in self._poset:
                if not self.has_meet(x, y) or not self.has_join(x, y):
                    return False
        return True

    def has_meet(self, x: T, y: T) -> bool:
        """Check if two elements have a greatest lower bound.

        Args:
            x: The first element.
            y: The second element.

        Returns:
            True if x and y have a greatest lower bound, False otherwise.

        """
        return self.get_meet(x, y) is not None

    def has_join(self, x: T, y: T) -> bool:
        """Check if two elements have a least upper bound.

        Args:
            x: The first element.
            y: The second element.

        Returns:
            True if x and y have a least upper bound, False otherwise.

        """
        return self.get_join(x, y) is not None

    def get_least_upper_bound(self, x: T, y: T) -> T:
        """Get the least upper bound of two elements.

        Args:
            x: The first element.
            y: The second element.

        Returns:
            The least upper bound of x and y.

        Raises:
            RuntimeError: If the least upper bound does not exist.

        """
        join = self.get_join(x, y)
        if join is None:
            raise RuntimeError(
                f"No least upper bound of {x} and {y} found for lattice."
            )
        return join

    def get_meet(self, x: T, y: T) -> T | None:
        """Get the greatest lower bound of two elements.

        Args:
            x: The first element.
            y: The second element.

        Returns:
            The greatest lower bound of x and y, or None if it does not exist.

        """
        meet = None
        for z in self._poset:
            if self._poset.is_less_than(z, x) and self._poset.is_less_than(z, y):
                if meet is None or self._poset.is_less_than(meet, z):
                    meet = z
        return meet

    def get_join(self, x: T, y: T) -> T | None:
        """Get the least upper bound of two elements.

        Args:
            x: The first element.
            y: The second element.

        Returns:
            The least upper bound of x and y, or None if it does not exist.

        """
        join = None
        for z in self._poset:
            if self._poset.is_greater_than(z, x) and self._poset.is_greater_than(z, y):
                if join is None or self._poset.is_greater_than(join, z):
                    join = z
        return join



"""Fundamental parameter classes."""

__all__ = [
    "NatParam",
    "is_the_basic_nat_param_constraint",
    "is_zero_included_constraint_exists",
    "is_zero_not_included_constraint_exists",
]

from collections.abc import Iterable
from typing import Any, cast

from fhy_core.constraint import Constraint, EquationConstraint
from fhy_core.expression import (
    BinaryExpression,
    BinaryOperation,
    IdentifierExpression,
    LiteralExpression,
)
from fhy_core.identifier import Identifier
from fhy_core.serialization import (
    DeserializationDictStructureError,
    DeserializationValueError,
    SerializedDict,
    register_serializable,
)
from fhy_core.utils import Self

from .core import (
    IntParam,
    ParamData,
    finalize_param_construction_from_data,
    is_valid_param_data,
)


def is_the_basic_nat_param_constraint(
    constraint: Constraint, variable: Identifier, is_zero_included: bool
) -> bool:
    """Return if the constraint is the basic constraint bounding the `NatParam`."""
    expression = constraint.convert_to_expression()
    if isinstance(expression, BinaryExpression) and (
        is_zero_included
        and expression.operation == BinaryOperation.GREATER_EQUAL
        or not is_zero_included
        and expression.operation == BinaryOperation.GREATER
    ):
        left = expression.left
        right = expression.right
        return (
            isinstance(left, IdentifierExpression)
            and left.identifier == variable
            and isinstance(right, LiteralExpression)
            and isinstance(right.value, int)
            and right.value == 0
        )
    return False


def is_zero_included_constraint_exists(
    constraints: Iterable[Constraint], variable: Identifier
) -> bool:
    """Return if there is a constraint bounding the variable to be GE to zero."""
    for constraint in constraints:
        expression = constraint.convert_to_expression()
        if (
            isinstance(expression, BinaryExpression)
            and expression.operation == BinaryOperation.GREATER_EQUAL
        ):
            left = expression.left
            right = expression.right
            if (
                isinstance(left, IdentifierExpression)
                and left.identifier == variable
                and isinstance(right, LiteralExpression)
                and isinstance(right.value, int)
                and right.value == 0
            ):
                return True
    return False


def is_zero_not_included_constraint_exists(
    constraints: Iterable[Constraint], variable: Identifier
) -> bool:
    """Return if there is a constraint bounding the variable to be GT to zero."""
    for constraint in constraints:
        expression = constraint.convert_to_expression()
        if (
            isinstance(expression, BinaryExpression)
            and expression.operation == BinaryOperation.GREATER
        ):
            left = expression.left
            right = expression.right
            if (
                isinstance(left, IdentifierExpression)
                and left.identifier == variable
                and isinstance(right, LiteralExpression)
                and isinstance(right.value, int)
                and right.value == 0
            ):
                return True
    return False


@register_serializable(type_id="nat_param")
class NatParam(IntParam):
    """Natural number parameter."""

    _is_zero_included: bool

    def __init__(
        self,
        *,
        name: Identifier | None = None,
        is_zero_included: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(name=name)
        object.__setattr__(self, "_is_zero_included", is_zero_included)
        if self._is_zero_included:
            basic_constraint = EquationConstraint(
                self.variable,
                BinaryExpression(
                    BinaryOperation.GREATER_EQUAL,
                    IdentifierExpression(self.variable),
                    LiteralExpression(0),
                ),
            )
        else:
            basic_constraint = EquationConstraint(
                self.variable,
                BinaryExpression(
                    BinaryOperation.GREATER,
                    IdentifierExpression(self.variable),
                    LiteralExpression(0),
                ),
            )
        object.__setattr__(
            self, "_constraints", self._constraints + (basic_constraint,)
        )

    def add_lower_bound_constraint(
        self, lower_bound: int, *, is_inclusive: bool = True
    ) -> "NatParam":
        if self._is_zero_included:
            if lower_bound < 0:
                raise ValueError("Lower bound must be non-negative.")
            if not is_inclusive and lower_bound < 1:
                raise ValueError(
                    "Lower bound must be at least 1 if zero is included and "
                    "bound is exclusive."
                )
        elif is_inclusive:
            if lower_bound < 1:
                raise ValueError(
                    "Lower bound must be at least 1 when zero is not included."
                )
        elif lower_bound < 0:
            raise ValueError(
                "Lower bound must be non-negative when zero is not included "
                "and bound is exclusive."
            )

        return super().add_lower_bound_constraint(
            lower_bound, is_inclusive=is_inclusive
        )

    def add_upper_bound_constraint(
        self, upper_bound: int, *, is_inclusive: bool = True
    ) -> "NatParam":
        if self._is_zero_included:
            if is_inclusive:
                if upper_bound < 0:
                    raise ValueError(
                        "Upper bound must be non-negative when zero is included."
                    )
            elif upper_bound < 1:
                raise ValueError(
                    "Upper bound must be at least 1 if zero is included and "
                    "bound is exclusive."
                )
        elif is_inclusive:
            if upper_bound < 1:
                raise ValueError(
                    "Upper bound must be at least 1 when zero is not included."
                )
        elif upper_bound < 2:  # noqa: PLR2004
            raise ValueError(
                "Upper bound must be at least 2 when zero is not included "
                "and bound is exclusive."
            )

        return super().add_upper_bound_constraint(
            upper_bound, is_inclusive=is_inclusive
        )

    def _clone(self) -> Self:
        new_param = self.__class__(
            name=self._variable, is_zero_included=self._is_zero_included
        )
        object.__setattr__(new_param, "_constraints", self._constraints)
        return new_param

    @classmethod
    def deserialize_data_from_dict(cls, data: SerializedDict) -> "NatParam":
        if not is_valid_param_data(data):
            raise DeserializationDictStructureError(
                cls, ParamData.__annotations__, data
            )

        variable = Identifier.deserialize_from_dict(data["variable"])
        constraints = [Constraint.deserialize_from_dict(c) for c in data["constraints"]]

        if is_zero_included_constraint_exists(constraints, variable):
            is_zero_included = True
        elif is_zero_not_included_constraint_exists(constraints, variable):
            is_zero_included = False
        else:
            raise DeserializationValueError(
                cls,
                "constraints",
                (
                    "a constraint bounding the variable to be greater than or "
                    "equal to zero or greater than zero"
                ),
                data["constraints"],
            )

        param = NatParam(name=variable, is_zero_included=is_zero_included)
        return cast(
            NatParam,
            finalize_param_construction_from_data(
                param,
                data,
                constraint_filter_function=lambda c: (
                    not is_the_basic_nat_param_constraint(c, variable, is_zero_included)
                ),
            ),
        )



"""Trait for IR nodes that can be visited by passes."""

__all__ = ["Visitable", "VisitableMixin"]

from abc import ABC
from typing import Any, Protocol


def _camel_to_snake(text: str) -> str:
    chars: list[str] = []
    for index, char in enumerate(text):
        if index > 0 and char.isupper():
            chars.append("_")
        chars.append(char.lower())
    return "".join(chars)


class _SupportsVisit(Protocol):
    def visit(self, node: "Visitable") -> Any: ...


class Visitable(ABC):
    """Interface for nodes that support pass visitor dispatch."""

    @classmethod
    def get_visit_method_suffix(cls) -> str:
        """Return visitor dispatch suffix for this node type."""
        return _camel_to_snake(cls.__name__)

    def accept(self, visitor: _SupportsVisit) -> Any:
        """Accept a visitor and return its visit result.

        Args:
            visitor: Visitor to accept, which must support visiting this node's type.

        Returns:
            The result of the visitor's visit method for this node.

        """
        return visitor.visit(self)


class VisitableMixin(Visitable):
    """Default `Visitable` behavior for IR node classes."""



"""Expression passes that interface with Z3."""

import operator
from typing import Any, Callable

import z3  # type: ignore
from frozendict import frozendict

from fhy_core.expression.core import (
    BinaryExpression,
    BinaryOperation,
    Expression,
    IdentifierExpression,
    LiteralExpression,
    SymbolType,
    UnaryExpression,
    UnaryOperation,
)
from fhy_core.identifier import Identifier
from fhy_core.pass_infrastructure import (
    PassExecutionError,
    VisitablePass,
    register_pass,
)


def _z3_floor_divide(left: z3.ExprRef, right: z3.ExprRef) -> z3.ExprRef:
    expr: z3.ArithRef = left / right
    if expr.is_real():
        return z3.ToInt(expr)
    elif expr.is_int():
        return expr
    else:
        raise ValueError(f"Unsupported floor divide expression type: {expr}")


@register_pass(
    "fhy_core.expression.to_z3",
    "Lower expression IR into an equivalent Z3 expression.",
)
class ExpressionToZ3Converter(VisitablePass[Expression, z3.ExprRef]):
    """Transforms an expression into a Z3 expression."""

    _UNARY_OPERATION_Z3_OPERATORS: frozendict[UnaryOperation, Callable[[Any], Any]] = (
        frozendict(
            {
                UnaryOperation.NEGATE: operator.neg,
                UnaryOperation.POSITIVE: operator.pos,
                UnaryOperation.LOGICAL_NOT: z3.Not,
            }
        )
    )
    _BINARY_OPERATION_Z3_OPERATORS: frozendict[
        BinaryOperation, Callable[[Any, Any], Any]
    ] = frozendict(
        {
            BinaryOperation.ADD: operator.add,
            BinaryOperation.SUBTRACT: operator.sub,
            BinaryOperation.MULTIPLY: operator.mul,
            BinaryOperation.DIVIDE: operator.truediv,
            BinaryOperation.FLOOR_DIVIDE: _z3_floor_divide,
            BinaryOperation.MODULO: operator.mod,
            BinaryOperation.POWER: operator.pow,
            BinaryOperation.LOGICAL_AND: z3.And,
            BinaryOperation.LOGICAL_OR: z3.Or,
            BinaryOperation.EQUAL: operator.eq,
            BinaryOperation.NOT_EQUAL: operator.ne,
            BinaryOperation.LESS: operator.lt,
            BinaryOperation.LESS_EQUAL: operator.le,
            BinaryOperation.GREATER: operator.gt,
            BinaryOperation.GREATER_EQUAL: operator.ge,
        }
    )

    _symbol_types: dict[Identifier, SymbolType]
    _identifier_to_z3_expression: dict[Identifier, z3.ExprRef]

    def __init__(self, symbol_types: dict[Identifier, SymbolType]) -> None:
        super().__init__()
        self._symbol_types = symbol_types
        self._identifier_to_z3_expression = {}

    @property
    def identifier_to_z3_expression(self) -> dict[Identifier, z3.ExprRef]:
        return self._identifier_to_z3_expression

    def visit_binary_expression(
        self, binary_expression: BinaryExpression
    ) -> z3.ExprRef:
        left = self.visit(binary_expression.left)
        right = self.visit(binary_expression.right)
        operator = self._BINARY_OPERATION_Z3_OPERATORS[binary_expression.operation]
        return operator(left, right)

    def visit_unary_expression(self, unary_expression: UnaryExpression) -> z3.ExprRef:
        operand = self.visit(unary_expression.operand)
        operator = self._UNARY_OPERATION_Z3_OPERATORS[unary_expression.operation]
        return operator(operand)

    def visit_identifier_expression(
        self, identifier_expression: IdentifierExpression
    ) -> z3.ExprRef:
        identifier_name = self.format_identifier(identifier_expression.identifier)
        identifier_type = self._symbol_types[identifier_expression.identifier]
        if identifier_type == SymbolType.REAL:
            result = z3.Real(identifier_name)
        elif identifier_type == SymbolType.INT:
            result = z3.Int(identifier_name)
        elif identifier_type == SymbolType.BOOL:
            result = z3.Bool(identifier_name)
        else:
            raise ValueError(f"Unsupported identifier type: {identifier_type}.")
        self._identifier_to_z3_expression[identifier_expression.identifier] = result
        return result

    def visit_literal_expression(
        self, literal_expression: LiteralExpression
    ) -> z3.ExprRef:
        if isinstance(literal_expression.value, bool):
            return z3.BoolVal(literal_expression.value)
        elif isinstance(literal_expression.value, int):
            return z3.IntVal(literal_expression.value)
        elif isinstance(literal_expression.value, float):
            return z3.RealVal(literal_expression.value)
        elif isinstance(literal_expression.value, str):
            if literal_expression.value == "True":
                return z3.BoolVal(True)
            elif literal_expression.value == "False":
                return z3.BoolVal(False)
            else:
                return z3.RealVal(literal_expression.value)
        else:
            raise TypeError(
                f"Unsupported literal type: {type(literal_expression.value)}"
            )

    @staticmethod
    def format_identifier(identifier: Identifier) -> str:
        return f"{identifier.name_hint}_{identifier.id}"

    def get_noop_output(self, ir: Expression) -> z3.ExprRef:
        raise PassExecutionError(
            f'Pass "{self.get_pass_name()}" does not define noop output.'
        )


def convert_expression_to_z3_expression(
    expression: Expression, symbol_types: dict[Identifier, SymbolType] | None = None
) -> tuple[z3.ExprRef, dict[Identifier, z3.ExprRef]]:
    """Convert an expression to a Z3 expression.

    Args:
        expression: Expression to convert.
        symbol_types: Symbol types.

    Returns:
        Z3 expression and mapping of identifiers to Z3 expressions.

    """
    converter = ExpressionToZ3Converter(symbol_types or {})
    z3_expression = converter(expression)
    return z3_expression, converter.identifier_to_z3_expression


def is_satisfiable(
    considered_identifiers: set[Identifier],
    expression: Expression,
    symbol_types: dict[Identifier, SymbolType],
) -> bool | None:
    """Check if the expression is satisfiable.

    Args:
        considered_identifiers: Considered identifiers.
        expression: Expression to check.
        symbol_types: Symbol types.

    Returns:
        True if the expression is satisfiable; False if the expression is unsatisfiable;
        None if the satisfiability is unknown.

    """
    z3_expression, identifier_to_z3_expression = convert_expression_to_z3_expression(
        expression, symbol_types
    )
    z3_expression = z3.Not(z3_expression)
    z3_expression = (
        z3.ForAll(
            [
                identifier_to_z3_expression[identifier]
                for identifier in considered_identifiers
            ],
            z3_expression,
        )
        if considered_identifiers
        else z3_expression
    )

    solver = z3.Solver()
    solver.add(z3_expression)

    result = solver.check()
    if result == z3.unsat:
        return True
    elif result == z3.sat:
        return False
    elif result == z3.unknown:
        return None
    else:
        raise RuntimeError("Unexpected Z3 result.")



"""Tests the core type system."""

import pytest
from fhy_core.expression import (
    IdentifierExpression,
    LiteralExpression,
)
from fhy_core.trait import Frozen, FrozenMutationError, StructuralEquivalence
from fhy_core.types import (
    CoreDataType,
    IndexType,
    NumericalType,
    PrimitiveDataType,
    TemplateDataType,
    TupleType,
    TypeQualifier,
    get_core_data_type_bit_width,
    promote_core_data_types,
    promote_type_qualifiers,
)

from .conftest import mock_identifier


def test_type_structural_equivalence_runtime_protocol():
    """Test `Type` implementations satisfy `StructuralEquivalence` protocol."""
    ty = NumericalType(PrimitiveDataType(CoreDataType.INT32))
    assert isinstance(ty, StructuralEquivalence)


def test_data_type_structural_equivalence_runtime_protocol():
    """Test `DataType` implementations satisfy `StructuralEquivalence` protocol."""
    data_type = PrimitiveDataType(CoreDataType.INT32)
    assert isinstance(data_type, StructuralEquivalence)


def test_type_family_is_frozen_on_construction():
    """Test all core type-family classes are frozen after construction."""
    N = mock_identifier("N", 1)
    shape = [IdentifierExpression(N), LiteralExpression(4)]
    data_type = PrimitiveDataType(CoreDataType.INT32)
    template_data_type = TemplateDataType(N, widths=[8, 16])
    numerical_type = NumericalType(data_type, shape)
    index_type = IndexType(LiteralExpression(0), LiteralExpression(10), None)
    tuple_type = TupleType([numerical_type, index_type])

    for value in (
        data_type,
        template_data_type,
        numerical_type,
        index_type,
        tuple_type,
    ):
        assert isinstance(value, Frozen)
        assert value.is_frozen
        with pytest.raises(FrozenMutationError):
            value._freeze_probe = "mutation"


def test_numerical_type_structural_equivalence_true():
    """Test structural equivalence is true for matching numerical types."""
    shape_1 = [LiteralExpression(4), LiteralExpression(8)]
    shape_2 = [LiteralExpression(4), LiteralExpression(8)]
    left = NumericalType(PrimitiveDataType(CoreDataType.INT32), shape_1)
    right = NumericalType(PrimitiveDataType(CoreDataType.INT32), shape_2)
    assert left.is_structurally_equivalent(right)


def test_numerical_type_structural_equivalence_false_for_data_type():
    """Test structural equivalence is false for differing numerical data types."""
    shape = [LiteralExpression(4)]
    left = NumericalType(PrimitiveDataType(CoreDataType.INT16), shape)
    right = NumericalType(PrimitiveDataType(CoreDataType.INT32), shape)
    assert not left.is_structurally_equivalent(right)


def test_index_type_structural_equivalence_false_for_stride():
    """Test structural equivalence is false for differing index stride values."""
    lower_bound = LiteralExpression(0)
    upper_bound = LiteralExpression(10)
    left = IndexType(lower_bound, upper_bound, LiteralExpression(1))
    right = IndexType(lower_bound, upper_bound, LiteralExpression(2))
    assert not left.is_structurally_equivalent(right)


def test_tuple_type_structural_equivalence_false_for_element_order():
    """Test structural equivalence is false for differing tuple type order."""
    int_type = NumericalType(PrimitiveDataType(CoreDataType.INT32))
    float_type = NumericalType(PrimitiveDataType(CoreDataType.FLOAT32))
    left = TupleType([int_type, float_type])
    right = TupleType([float_type, int_type])
    assert not left.is_structurally_equivalent(right)


@pytest.mark.parametrize(
    "core_data_type, expected_bit_width",
    [
        (CoreDataType.UINT8, 8),
        (CoreDataType.UINT16, 16),
        (CoreDataType.UINT32, 32),
        (CoreDataType.UINT64, 64),
        (CoreDataType.INT8, 8),
        (CoreDataType.INT16, 16),
        (CoreDataType.INT32, 32),
        (CoreDataType.INT64, 64),
        (CoreDataType.FLOAT16, 16),
        (CoreDataType.FLOAT32, 32),
        (CoreDataType.FLOAT64, 64),
        (CoreDataType.COMPLEX32, 32),
        (CoreDataType.COMPLEX64, 64),
        (CoreDataType.COMPLEX128, 128),
    ],
)
def test_get_core_data_type_bit_width(core_data_type, expected_bit_width):
    """Test get_core_data_type_bit_width function with various core data types."""
    assert get_core_data_type_bit_width(core_data_type) == expected_bit_width


@pytest.mark.parametrize(
    ("core_data_type1", "core_data_type2", "expected_core_data_type"),
    [
        (CoreDataType.UINT8, CoreDataType.UINT8, CoreDataType.UINT8),
        (CoreDataType.UINT8, CoreDataType.UINT16, CoreDataType.UINT16),
        (CoreDataType.UINT16, CoreDataType.UINT8, CoreDataType.UINT16),
        (CoreDataType.INT32, CoreDataType.INT64, CoreDataType.INT64),
        (
            CoreDataType.FLOAT16,
            CoreDataType.FLOAT32,
            CoreDataType.FLOAT32,
        ),
        (
            CoreDataType.FLOAT64,
            CoreDataType.FLOAT16,
            CoreDataType.FLOAT64,
        ),
        (
            CoreDataType.COMPLEX32,
            CoreDataType.COMPLEX64,
            CoreDataType.COMPLEX64,
        ),
        (
            CoreDataType.FLOAT32,
            CoreDataType.COMPLEX32,
            CoreDataType.COMPLEX64,
        ),
    ],
)
def test_promote_primitive_data_type(
    core_data_type1, core_data_type2, expected_core_data_type
):
    """Test primitive data types are correctly promoted."""
    assert (
        promote_core_data_types(core_data_type1, core_data_type2)
        == expected_core_data_type
    ), (
        f"Expected the promotion of {core_data_type1} and {core_data_type2} "
        f"to be {expected_core_data_type}."
    )


@pytest.mark.parametrize(
    ("type_qualifer1", "type_qualifer2", "expected_type_qualifer"),
    [
        (TypeQualifier.INPUT, TypeQualifier.INPUT, TypeQualifier.TEMP),
        (TypeQualifier.STATE, TypeQualifier.PARAM, TypeQualifier.TEMP),
        (TypeQualifier.PARAM, TypeQualifier.TEMP, TypeQualifier.TEMP),
        (TypeQualifier.PARAM, TypeQualifier.PARAM, TypeQualifier.PARAM),
    ],
)
def test_promote_type_qualifiers(
    type_qualifer1, type_qualifer2, expected_type_qualifer
):
    """Test type qualifiers are correctly promoted."""
    assert (
        promote_type_qualifiers(type_qualifer1, type_qualifer2)
        == expected_type_qualifer
    ), (
        f"Expected the promotion of {type_qualifer1} and {type_qualifer2} "
        f"to be {expected_type_qualifer}."
    )


def test_primitive_data_type_dict_serialization():
    """Test primitive data types can be serialized/deserialized via a dictionary."""
    for core_data_type in CoreDataType:
        primitive_data_type = PrimitiveDataType(core_data_type)
        expected_dict = {
            "__type__": "primitive_data_type",
            "__data__": {"core_data_type": core_data_type.value},
        }
        dictionary = primitive_data_type.serialize_to_dict()
        assert dictionary == expected_dict
        primitive_data_type_deserialized = PrimitiveDataType.deserialize_from_dict(
            dictionary
        )
        assert isinstance(primitive_data_type_deserialized, PrimitiveDataType)
        assert primitive_data_type_deserialized.core_data_type == core_data_type


def test_numerical_type_dict_serialization():
    """Test numerical types can be serialized/deserialized via a dictionary."""
    N = mock_identifier("N", 1)
    shape = [
        IdentifierExpression(N),
        LiteralExpression(28),
    ]
    numerical_type = NumericalType(PrimitiveDataType(CoreDataType.INT32), shape)
    expected_dict = {
        "__type__": "numerical_type",
        "__data__": {
            "data_type": {
                "__type__": "primitive_data_type",
                "__data__": {"core_data_type": CoreDataType.INT32.value},
            },
            "shape": [
                shape[0].serialize_to_dict(),
                shape[1].serialize_to_dict(),
            ],
        },
    }
    dictionary = numerical_type.serialize_to_dict()
    assert dictionary == expected_dict
    numerical_type_deserialized = NumericalType.deserialize_from_dict(dictionary)
    assert isinstance(numerical_type_deserialized, NumericalType)
    assert isinstance(numerical_type_deserialized.data_type, PrimitiveDataType)
    assert numerical_type_deserialized.data_type.core_data_type == CoreDataType.INT32
    assert len(numerical_type_deserialized.shape) == 2
    assert numerical_type_deserialized.shape[0].is_structurally_equivalent(shape[0])
    assert numerical_type_deserialized.shape[1].is_structurally_equivalent(shape[1])


def test_index_type_dict_serialization():
    """Test index types can be serialized/deserialized via a dictionary."""
    N = mock_identifier("N", 1)
    lower_bound = LiteralExpression(1)
    upper_bound = IdentifierExpression(N)
    index_type = IndexType(lower_bound, upper_bound, None)
    expected_dict = {
        "__type__": "index_type",
        "__data__": {
            "lower_bound": lower_bound.serialize_to_dict(),
            "upper_bound": upper_bound.serialize_to_dict(),
            "stride": None,
        },
    }
    dictionary = index_type.serialize_to_dict()
    assert dictionary == expected_dict
    index_type_deserialized = IndexType.deserialize_from_dict(dictionary)
    assert isinstance(index_type_deserialized, IndexType)
    assert index_type_deserialized.lower_bound.is_structurally_equivalent(lower_bound)
    assert index_type_deserialized.upper_bound.is_structurally_equivalent(upper_bound)


def test_index_type_with_stride_serialization():
    """Test index types with stride can be serialized/deserialized via a dictionary."""
    N = mock_identifier("N", 1)
    lower_bound = LiteralExpression(1)
    upper_bound = IdentifierExpression(N)
    stride = LiteralExpression(2)
    index_type = IndexType(lower_bound, upper_bound, stride)
    expected_dict = {
        "__type__": "index_type",
        "__data__": {
            "lower_bound": lower_bound.serialize_to_dict(),
            "upper_bound": upper_bound.serialize_to_dict(),
            "stride": stride.serialize_to_dict(),
        },
    }
    dictionary = index_type.serialize_to_dict()
    assert dictionary == expected_dict
    index_type_deserialized = IndexType.deserialize_from_dict(dictionary)
    assert isinstance(index_type_deserialized, IndexType)
    assert index_type_deserialized.lower_bound.is_structurally_equivalent(lower_bound)
    assert index_type_deserialized.upper_bound.is_structurally_equivalent(upper_bound)
    assert index_type_deserialized.stride is not None
    assert index_type_deserialized.stride.is_structurally_equivalent(stride)


def test_tuple_type_dict_serialization():
    """Test tuple types can be serialized/deserialized via a dictionary."""
    N = mock_identifier("N", 1)
    shape = [
        IdentifierExpression(N),
        LiteralExpression(28),
    ]
    numerical_type = NumericalType(PrimitiveDataType(CoreDataType.INT32), shape)
    tuple_type = TupleType([numerical_type, numerical_type])
    expected_dict = {
        "__type__": "tuple_type",
        "__data__": {
            "types": [
                numerical_type.serialize_to_dict(),
                numerical_type.serialize_to_dict(),
            ],
        },
    }
    dictionary = tuple_type.serialize_to_dict()
    assert dictionary == expected_dict
    tuple_type_deserialized = TupleType.deserialize_from_dict(dictionary)
    assert isinstance(tuple_type_deserialized, TupleType)
    assert len(tuple_type_deserialized.types) == 2
    for ty in tuple_type_deserialized.types:
        assert isinstance(ty, NumericalType)
        assert isinstance(ty.data_type, PrimitiveDataType)
        assert ty.data_type.core_data_type == CoreDataType.INT32
        assert len(ty.shape) == 2
        assert ty.shape[0].is_structurally_equivalent(shape[0])
        assert ty.shape[1].is_structurally_equivalent(shape[1])


# TODO: Check serialization structure errors and value errors for all types.



"""General expression tree utility."""

__all__ = [
    "BinaryExpression",
    "BinaryOperation",
    "Expression",
    "IdentifierExpression",
    "LiteralExpression",
    "LiteralType",
    "UnaryExpression",
    "UnaryOperation",
    "collect_identifiers",
    "convert_expression_to_sympy_expression",
    "convert_expression_to_z3_expression",
    "convert_sympy_expression_to_expression",
    "is_satisfiable",
    "parse_expression",
    "pformat_expression",
    "replace_identifiers",
    "simplify_expression",
    "substitute_identifiers",
    "substitute_sympy_expression_variables",
    "SymbolType",
    "tokenize_expression",
]

from .core import (
    BinaryExpression,
    BinaryOperation,
    Expression,
    IdentifierExpression,
    LiteralExpression,
    LiteralType,
    SymbolType,
    UnaryExpression,
    UnaryOperation,
)
from .parser import parse_expression, tokenize_expression
from .passes import (
    collect_identifiers,
    convert_expression_to_sympy_expression,
    convert_expression_to_z3_expression,
    convert_sympy_expression_to_expression,
    is_satisfiable,
    replace_identifiers,
    simplify_expression,
    substitute_identifiers,
    substitute_sympy_expression_variables,
)
from .pprint import pformat_expression



"""Tests the serialization module."""

import importlib
import json
import struct
import types
from abc import ABC
from dataclasses import dataclass
from typing import Any, Mapping

import pytest
from fhy_core import serialization


@serialization.register_serializable
@dataclass(frozen=True)
class _DummySpan(serialization.Serializable):
    lo: int
    hi: int

    def serialize_to_dict(self) -> dict[str, Any]:
        return {"lo": self.lo, "hi": self.hi}

    @classmethod
    def deserialize_from_dict(cls, data: Mapping[str, Any]) -> "_DummySpan":
        return cls(int(data["lo"]), int(data["hi"]))


@serialization.register_serializable
@dataclass(frozen=True)
class _CompactSpan(serialization.Serializable):
    """A span that uses a compact custom binary payload (8 bytes)."""

    lo: int
    hi: int

    def serialize_to_dict(self) -> dict[str, Any]:
        return {"lo": self.lo, "hi": self.hi}

    @classmethod
    def deserialize_from_dict(cls, data: Mapping[str, Any]) -> "_CompactSpan":
        return cls(int(data["lo"]), int(data["hi"]))

    @classmethod
    def get_binary_codec(cls) -> serialization.BinaryPayloadCodec:
        return serialization.BinaryPayloadCodec.CUSTOM

    def serialize_to_binary(self, *, codec: serialization.BinaryPayloadCodec) -> bytes:
        if codec is not serialization.BinaryPayloadCodec.CUSTOM:
            raise serialization.CodecMismatchError("Expected CUSTOM codec.")
        return struct.pack("!ii", self.lo, self.hi)

    @classmethod
    def deserialize_from_binary(
        cls,
        payload: bytes,
        *,
        codec: serialization.BinaryPayloadCodec,
    ) -> "_CompactSpan":
        if codec is not serialization.BinaryPayloadCodec.CUSTOM:
            raise serialization.CodecMismatchError("Expected CUSTOM codec.")
        lo, hi = struct.unpack("!ii", payload)
        return cls(lo=lo, hi=hi)


@serialization.register_serializable(type_id="compiler_core.CustomIdNode")
@dataclass(frozen=True)
class _CustomIdNode(serialization.Serializable):
    x: int

    TYPE_ID = "compiler_core.CustomIdNode"

    def serialize_to_dict(self) -> dict[str, Any]:
        return {"x": self.x}

    @classmethod
    def deserialize_from_dict(cls, data: Mapping[str, Any]) -> "_CustomIdNode":
        return cls(int(data["x"]))


class _CustomNodeBase(serialization.WrappedFamilySerializable, ABC):
    pass


@serialization.register_serializable
class _CustomNode(_CustomNodeBase):
    value: int

    def __init__(self, value: int):
        self.value = value

    def serialize_data_to_dict(self) -> dict[str, Any]:
        return {"value": self.value}

    @classmethod
    def deserialize_data_from_dict(cls, data: Mapping[str, Any]) -> "_CustomNode":
        return cls(data["value"])


def test_register_serializable_refuses_override_by_default():
    """Test registering a different class under an existing type_id is rejected."""

    @serialization.register_serializable(type_id="test.DuplicateId")
    class _First(serialization.Serializable):
        def serialize_to_dict(self) -> dict[str, Any]:
            return {"x": 1}

        @classmethod
        def deserialize_from_dict(cls, data: Mapping[str, Any]) -> "_First":
            return cls()

    with pytest.raises(serialization.SerializationError):

        @serialization.register_serializable(type_id="test.DuplicateId")
        class _Second(serialization.Serializable):
            def serialize_to_dict(self) -> dict[str, Any]:
                return {"x": 2}

            @classmethod
            def deserialize_from_dict(cls, data: Mapping[str, Any]) -> "_Second":
                return cls()


def test_register_serializable_canonical_conflict_rejected():
    """Test registering a class with a conflicting canonical type_id is rejected."""

    @serialization.register_serializable(type_id="test.CanonicalA")
    class _Canon(serialization.Serializable):
        def serialize_to_dict(self) -> dict[str, Any]:
            return {}

        @classmethod
        def deserialize_from_dict(cls, data: Mapping[str, Any]) -> "_Canon":
            return cls()

    with pytest.raises(serialization.SerializationError):
        serialization.register_serializable(_Canon, type_id="test.CanonicalB")


def test_register_serializable_allows_alias_without_changing_canonical_id():
    """Test alias registration works without changing the class' canonical type id."""

    @serialization.register_serializable(type_id="test.CanonicalMain")
    class _Aliased(serialization.Serializable):
        def serialize_to_dict(self) -> dict[str, Any]:
            return {"ok": True}

        @classmethod
        def deserialize_from_dict(cls, data: Mapping[str, Any]) -> "_Aliased":
            return cls()

    serialization.register_serializable(
        _Aliased, type_id="test.LegacyAlias", alias=True
    )

    assert _Aliased.get_serialization_class_type_id() == "test.CanonicalMain"
    assert serialization._resolve_type_id("test.CanonicalMain") is _Aliased
    assert serialization._resolve_type_id("test.LegacyAlias") is _Aliased


def test_binary_deserialization_accepts_alias_type_id():
    """Test binary deserialization accepts a legacy alias type_id in the envelope."""
    span = _DummySpan(11, 22)
    blob = span.to_bytes()
    magic, version, codec_u8, _, payload = _parse_envelope(blob)

    legacy_id = "test.LegacyDummySpan"
    serialization.register_serializable(_DummySpan, type_id=legacy_id, alias=True)

    header_struct = struct.Struct("!4sBBH")
    payload_len_struct = struct.Struct("!I")

    rebuilt = (
        header_struct.pack(magic, version, codec_u8, len(legacy_id.encode("utf-8")))
        + legacy_id.encode("utf-8")
        + payload_len_struct.pack(len(payload))
        + payload
    )

    obj = serialization.Serializable.from_bytes(rebuilt)
    assert isinstance(obj, _DummySpan)
    assert obj == span


def test_wrapped_family_serializable_accepts_alias_type_id():
    """Test `WrappedFamilySerializable` accepts a legacy alias via `__type__`."""
    legacy_id = "test.LegacyCustomNode"
    serialization.register_serializable(_CustomNode, type_id=legacy_id, alias=True)

    wrapped = {"__type__": legacy_id, "__data__": {"value": 7}}

    node2 = _CustomNodeBase.deserialize(wrapped, serialization.SerializationFormat.DICT)
    assert isinstance(node2, _CustomNode)
    assert node2.value == 7


def test_dict_round_trip_via_classmethod():
    """Test that `serialize`/`deserialize` via dict format works."""
    span = _DummySpan(1, 9)
    dictionary = span.serialize(serialization.SerializationFormat.DICT)
    assert dictionary == {"lo": 1, "hi": 9}

    span2 = _DummySpan.deserialize(dictionary, serialization.SerializationFormat.DICT)
    assert span2 == span


def test_json_round_trip_via_classmethod():
    """Test that `serialize`/`deserialize` via JSON format works."""
    span = _DummySpan(2, 7)
    json_str = span.serialize(serialization.SerializationFormat.JSON)
    assert isinstance(json_str, str)

    span2 = _DummySpan.deserialize(json_str, serialization.SerializationFormat.JSON)
    assert span2 == span

    span3 = _DummySpan.deserialize(
        json_str.encode("utf-8"), serialization.SerializationFormat.JSON
    )
    assert span3 == span


def test_serialize_dispatch_and_invalid_format():
    """Test that `serialize` returns the correct format/rejects invalid formats."""
    span = _DummySpan(0, 1)
    assert span.serialize(serialization.SerializationFormat.DICT) == {"lo": 0, "hi": 1}
    assert isinstance(span.serialize(serialization.SerializationFormat.JSON), str)
    assert isinstance(
        span.serialize(serialization.SerializationFormat.BINARY), (bytes, bytearray)
    )

    with pytest.raises(serialization.SerializationError):
        span.serialize("nope")


def test_deserialize_type_mismatch_raises():
    """Test that deserializing with a type that doesn't match raises an error."""
    blob = _DummySpan(1, 2).to_bytes()
    with pytest.raises(serialization.SerializationError):
        _CustomIdNode.deserialize(blob, serialization.SerializationFormat.BINARY)


def test_wrapped_family_serializable_round_trip():
    """Test that a `WrappedFamilySerializable` can round-trip."""
    node = _CustomNode(42)

    dictionary = node.serialize(serialization.SerializationFormat.DICT)
    assert dictionary == {
        "__type__": node.get_serialization_class_type_id(),
        "__data__": {"value": 42},
    }

    node2 = _CustomNodeBase.deserialize(
        dictionary, serialization.SerializationFormat.DICT
    )
    assert isinstance(node2, _CustomNode)
    assert node2.value == 42

    node3 = _CustomNode.deserialize(dictionary, serialization.SerializationFormat.DICT)
    assert isinstance(node3, _CustomNode)
    assert node3.value == 42


def _parse_envelope(blob: bytes) -> tuple[bytes, int, int, str, bytes]:
    header_struct = struct.Struct("!4sBBH")
    payload_len_struct = struct.Struct("!I")

    magic, version, codec_u8, type_id_len = header_struct.unpack(
        blob[: header_struct.size]
    )
    offset = header_struct.size

    type_id = blob[offset : offset + type_id_len].decode("utf-8")
    offset += type_id_len

    (payload_len,) = payload_len_struct.unpack(
        blob[offset : offset + payload_len_struct.size]
    )
    offset += payload_len_struct.size

    payload = blob[offset : offset + payload_len]
    return magic, version, codec_u8, type_id, payload


def test_binary_round_trip_auto_reconstructs_class_default_json_codec():
    """Test that serializing to binary and deserializing reconstructs the object."""
    span = _DummySpan(10, 20)
    blob = span.to_bytes()

    obj = serialization.Serializable.from_bytes(blob)
    assert isinstance(obj, _DummySpan)
    assert obj == span


def test_binary_round_trip_auto_reconstructs_class_custom_compact_codec():
    """Test that custom compact binary payload round-trips and auto-reconstructs."""
    span = _CompactSpan(10, 20)
    blob = span.to_bytes()

    obj = serialization.Serializable.from_bytes(blob)
    assert isinstance(obj, _CompactSpan)
    assert obj == span


def test_binary_envelope_contains_expected_fields_json_codec():
    """Test that the envelope contains the expected fields for JSON codec payloads."""
    span = _DummySpan(5, 6)
    blob = span.to_bytes()

    magic, version, codec_u8, type_id, payload = _parse_envelope(blob)
    assert magic == b"FhYS"
    assert version == 1
    assert codec_u8 == 1  # JSON
    assert type_id == _DummySpan.get_serialization_class_type_id()

    payload_obj = json.loads(payload.decode("utf-8"))
    assert payload_obj == {"lo": 5, "hi": 6}


def test_binary_envelope_contains_expected_fields_custom_codec():
    """Test that the envelope contains the expected fields for CUSTOM codec payloads."""
    span = _CompactSpan(5, 6)
    blob = span.to_bytes()

    magic, version, codec_u8, type_id, payload = _parse_envelope(blob)
    assert magic == b"FhYS"
    assert version == 1
    assert codec_u8 == 2  # CUSTOM
    assert type_id == _CompactSpan.get_serialization_class_type_id()

    lo, hi = struct.unpack("!ii", payload)
    assert (lo, hi) == (5, 6)


def test_binary_custom_type_id_is_used():
    """Test that a custom type id is used in the binary envelope."""
    node = _CustomIdNode(123)
    blob = node.to_bytes()
    _, _, codec_u8, type_id, payload = _parse_envelope(blob)

    assert codec_u8 == 1  # default JSON codec
    assert type_id == "compiler_core.CustomIdNode"
    assert json.loads(payload.decode("utf-8")) == {"x": 123}

    obj = serialization.Serializable.from_bytes(blob)
    assert isinstance(obj, _CustomIdNode)
    assert obj == node


def test_binary_bad_magic_raises():
    """Test that a blob with a bad magic value raises an error."""
    blob = _DummySpan(1, 2).to_bytes()
    corrupted = b"NOPE" + blob[4:]
    with pytest.raises(serialization.SerializationError):
        serialization.Serializable.from_bytes(corrupted)


def test_binary_version_mismatch_raises():
    """Test that a blob with an unsupported version raises a `VersionMismatchError`."""
    blob = _DummySpan(1, 2).to_bytes()

    corrupted = bytearray(blob)
    corrupted[4] = 99  # version byte
    with pytest.raises(serialization.VersionMismatchError):
        serialization.Serializable.from_bytes(bytes(corrupted))


def test_binary_unknown_codec_raises():
    """Test that a blob with an unknown codec code raises an `UnknownCodecError`."""
    blob = _DummySpan(1, 2).to_bytes()
    corrupted = bytearray(blob)
    corrupted[5] = 250  # codec byte (unknown)
    with pytest.raises(serialization.UnknownCodecError):
        serialization.Serializable.from_bytes(bytes(corrupted))


def test_binary_codec_mismatch_raises_on_custom_class_with_json_envelope():
    """Test that codec mismatch is rejected (CUSTOM class, JSON envelope)."""
    span = _CompactSpan(1, 2)
    blob = span.to_bytes()
    corrupted = bytearray(blob)
    corrupted[5] = 1  # force JSON codec in envelope
    with pytest.raises(serialization.CodecMismatchError):
        serialization.Serializable.from_bytes(bytes(corrupted))


def test_binary_truncated_header_raises():
    """Test that a blob with a truncated header raises an error."""
    blob = _DummySpan(1, 2).to_bytes()
    with pytest.raises(serialization.SerializationError):
        serialization.Serializable.from_bytes(blob[:2])


def test_binary_truncated_type_id_or_payload_raises():
    """Test that a blob with a truncated type id or payload raises an error."""
    blob = _DummySpan(1, 2).to_bytes()
    with pytest.raises(serialization.SerializationError):
        serialization.Serializable.from_bytes(blob[:-1])


def test_binary_unknown_type_id_raises():
    """Test that a blob with an unknown type id raises an `UnknownTypeIdError`."""
    blob = _DummySpan(1, 2).to_bytes()
    magic, version, codec_u8, _, payload = _parse_envelope(blob)

    header_struct = struct.Struct("!4sBBH")
    payload_len_struct = struct.Struct("!I")

    fake_type_id = "does.not.Exist".encode("utf-8")
    rebuilt = (
        header_struct.pack(magic, version, codec_u8, len(fake_type_id))
        + fake_type_id
        + payload_len_struct.pack(len(payload))
        + payload
    )

    with pytest.raises(serialization.UnknownTypeIdError):
        serialization.Serializable.from_bytes(rebuilt)


def test_registry_resolution_prefers_registered_class():
    """Test that the registry resolution prefers registered classes over importlib."""
    cls = serialization._resolve_type_id(_DummySpan.get_serialization_class_type_id())
    assert cls is _DummySpan


def test_fallback_import_resolution_via_mock(monkeypatch):
    """Test that import fallback works when explicitly enabled and allowlisted."""
    mod = types.SimpleNamespace()

    class _Outer(serialization.Serializable):
        def serialize_to_dict(self) -> dict[str, Any]:
            return {"ok": True}

        @classmethod
        def deserialize_from_dict(cls, data: Mapping[str, Any]) -> "_Outer":
            return cls()

    setattr(mod, "_Outer", _Outer)

    def fake_import(name: str):
        assert name == "fakepkg.fakemod"
        return mod

    monkeypatch.setattr(importlib, "import_module", fake_import)

    ty_id = "fakepkg.fakemod._Outer"

    resolved = serialization._resolve_type_id(
        ty_id,
        allow_import_fallback=True,
        allowed_module_prefixes=("fakepkg.",),
    )
    assert resolved is _Outer


def test_fallback_import_resolution_disabled_by_default(monkeypatch):
    """Test that import fallback is disabled by default (registry-only)."""

    def fail_import(name: str):
        raise AssertionError("importlib.import_module should not be called by default")

    monkeypatch.setattr(importlib, "import_module", fail_import)

    with pytest.raises(serialization.UnknownTypeIdError):
        serialization._resolve_type_id("fakepkg.fakemod._Outer")


def test_fallback_import_resolution_rejects_non_serializable(monkeypatch):
    """Test that the fallback resolution rejects types that are not serializable."""
    mod = types.SimpleNamespace()

    class _NotSerializable:
        pass

    setattr(mod, "_NotSerializable", _NotSerializable)

    def fake_import(name: str):
        assert name == "fakepkg.fakemod"
        return mod

    monkeypatch.setattr(importlib, "import_module", fake_import)

    with pytest.raises(serialization.UnknownTypeIdError):
        serialization._resolve_type_id("fakepkg.fakemod._NotSerializable")


def test_binary_deserialization_does_not_import_by_default(monkeypatch):
    """Test by default, deserialization does not import unknown types."""
    blob = _DummySpan(1, 2).to_bytes()
    magic, version, codec_u8, _, payload = _parse_envelope(blob)

    fake_type_id = "fakepkg.fakemod._Outer".encode("utf-8")
    header_struct = struct.Struct("!4sBBH")
    payload_len_struct = struct.Struct("!I")

    rebuilt = (
        header_struct.pack(magic, version, codec_u8, len(fake_type_id))
        + fake_type_id
        + payload_len_struct.pack(len(payload))
        + payload
    )

    def fail_import(name: str):
        raise AssertionError(
            "importlib.import_module was called but fallback should be disabled"
        )

    monkeypatch.setattr(importlib, "import_module", fail_import)

    with pytest.raises(serialization.UnknownTypeIdError):
        serialization.Serializable.from_bytes(rebuilt)


def test_binary_deserialization_import_fallback_opt_in(monkeypatch):
    """Test deserialization can import/deserialize a type not in the registry."""
    blob = _DummySpan(1, 2).to_bytes()
    magic, version, codec_u8, _, payload = _parse_envelope(blob)

    mod = types.SimpleNamespace()

    class _Outer(serialization.Serializable):
        def serialize_to_dict(self) -> dict[str, Any]:
            return {"ok": True}

        @classmethod
        def deserialize_from_dict(cls, data: Mapping[str, Any]) -> "_Outer":
            return cls()

    setattr(mod, "_Outer", _Outer)

    def fake_import(name: str):
        assert name == "fakepkg.fakemod"
        return mod

    monkeypatch.setattr(importlib, "import_module", fake_import)

    fake_type_id = "fakepkg.fakemod._Outer".encode("utf-8")
    header_struct = struct.Struct("!4sBBH")
    payload_len_struct = struct.Struct("!I")

    rebuilt = (
        header_struct.pack(magic, version, codec_u8, len(fake_type_id))
        + fake_type_id
        + payload_len_struct.pack(len(payload))
        + payload
    )

    obj = serialization.Serializable.from_bytes(
        rebuilt, allow_import_fallback=True, allowed_module_prefixes=("fakepkg.",)
    )
    assert isinstance(obj, _Outer)


def test_from_json_non_object_raises():
    """Test that from_json raises if the JSON is not an object."""
    with pytest.raises(serialization.SerializationError):
        _DummySpan.from_json("[]")


def test_deserialize_invalid_payload_type_raises():
    """Test that deserializing with an invalid payload type raises an error."""
    with pytest.raises(serialization.SerializationError):
        _DummySpan.deserialize(123, serialization.SerializationFormat.DICT)
    with pytest.raises(serialization.SerializationError):
        _DummySpan.deserialize(123, serialization.SerializationFormat.JSON)
    with pytest.raises(serialization.SerializationError):
        _DummySpan.deserialize("not-bytes", serialization.SerializationFormat.BINARY)


def test_registry_wrapped_value_round_trip_for_int():
    """Test wrapped-value helpers round-trip integer values."""
    wrapped = serialization.serialize_registry_wrapped_value(7)
    restored = serialization.deserialize_registry_wrapped_value(wrapped)
    assert restored == 7


def test_registry_wrapped_value_round_trip_for_bool():
    """Test wrapped-value helpers round-trip boolean values."""
    wrapped = serialization.serialize_registry_wrapped_value(True)
    restored = serialization.deserialize_registry_wrapped_value(wrapped)
    assert restored is True


def test_registry_wrapped_value_round_trip_for_string():
    """Test wrapped-value helpers round-trip string values."""
    wrapped = serialization.serialize_registry_wrapped_value("value")
    restored = serialization.deserialize_registry_wrapped_value(wrapped)
    assert restored == "value"


def test_registry_wrapped_value_round_trip_for_serializable():
    """Test wrapped-value helpers round-trip serializable values."""
    value = _DummySpan(4, 9)
    wrapped = serialization.serialize_registry_wrapped_value(value)
    restored = serialization.deserialize_registry_wrapped_value(wrapped)
    assert isinstance(restored, _DummySpan)
    assert restored == value


def test_registry_wrapped_value_round_trip_for_tuple():
    """Test wrapped-value helpers round-trip tuple values."""
    wrapped = serialization.serialize_registry_wrapped_value((1, "a", True))
    restored = serialization.deserialize_registry_wrapped_value(wrapped)
    assert restored == (1, "a", True)


def test_registry_wrapped_value_round_trip_for_frozenset():
    """Test wrapped-value helpers round-trip frozenset values."""
    wrapped = serialization.serialize_registry_wrapped_value(frozenset({1, 2, 3}))
    restored = serialization.deserialize_registry_wrapped_value(wrapped)
    assert restored == frozenset({1, 2, 3})


def test_registry_wrapped_value_round_trip_for_nested_collections():
    """Test wrapped-value helpers round-trip nested tuple/frozenset values."""
    value = (1, (2, 3), frozenset({4, 5}), _DummySpan(6, 7))
    wrapped = serialization.serialize_registry_wrapped_value(value)
    restored = serialization.deserialize_registry_wrapped_value(wrapped)
    assert restored == value


def test_registry_wrapped_value_deserialization_rejects_bad_structure():
    """Test wrapped-value deserialization rejects invalid wrapped dicts."""
    with pytest.raises(serialization.DeserializationDictStructureError):
        serialization.deserialize_registry_wrapped_value({"bad": "data"})



"""Tests the parameter utility."""

from functools import partial
from typing import TypeVar

import pytest
from fhy_core.constraint import EquationConstraint, InSetConstraint
from fhy_core.expression import SymbolType
from fhy_core.param import (
    BoundIntParam,
    BoundNatParam,
    CategoricalParam,
    IntParam,
    NatParam,
    OrdinalParam,
    Param,
    ParamAssignment,
    PermParam,
    RealParam,
)

T = TypeVar("T")


def _assert_all_satisfied(param: Param[T], values: list[T]) -> None:
    for v in values:
        assert param.is_constraints_satisfied(
            v
        ), f"Value {v} should satisfy constraints of parameter {param}"


def _assert_none_satisfied(param: Param[T], values: list[T]) -> None:
    for v in values:
        assert not param.is_constraints_satisfied(
            v
        ), f"Value {v} should not satisfy constraints of parameter {param}"


@pytest.fixture
def default_real_param() -> RealParam:
    return RealParam()


@pytest.fixture
def default_int_param() -> IntParam:
    return IntParam()


def test_param_is_not_set_after_initialization(default_real_param):
    """Test that the value of a parameter is not set after initialization."""
    assignment = default_real_param.set_value(1.0)
    assert isinstance(default_real_param, RealParam)
    assert isinstance(assignment, ParamAssignment)
    assert assignment.param is default_real_param
    assert not hasattr(default_real_param, "is_value_set")


def test_param_is_set_after_setting_value(default_real_param):
    """Test that the value of a parameter is set after setting a value."""
    default_real_param = default_real_param.set_value(1.0)
    assert default_real_param.is_value_set()


def test_param_get_value_fails_when_not_set(default_real_param):
    """Test that parameters no longer expose direct value access."""
    with pytest.raises(AttributeError):
        default_real_param.get_value()


def test_get_param_value_after_setting_value(default_real_param):
    """Test that the value of a parameter can be retrieved after setting a value."""
    assignment = default_real_param.set_value(1.0)
    assert assignment.value == 1.0


def test_real_param_get_symbol_type(default_real_param):
    """Test that the symbol type of a real parameter is correct."""
    assert default_real_param.get_symbol_type() == SymbolType.REAL


def test_int_param_get_symbol_type(default_int_param):
    """Test that the symbol type of an integer parameter is correct."""
    assert default_int_param.get_symbol_type() == SymbolType.INT


def test_real_param_value_set_fails_with_invalid_value(default_real_param):
    """Test that setting a real parameter value fails with an invalid value."""
    with pytest.raises(ValueError):
        default_real_param = default_real_param.set_value([])


def test_int_param_value_set_fails_with_invalid_value(default_int_param):
    """Test that setting an integer parameter value fails with an invalid value."""
    with pytest.raises(ValueError):
        default_int_param = default_int_param.set_value(1.0)


def test_create_real_param_with_value():
    """Test that a real parameter can be created with a value."""
    param = RealParam.with_value(1.0)
    assert param.is_value_set()
    assert param.value == 1.0


def test_create_real_param_with_value_fails_with_invalid_value():
    """Test that creating a real parameter with an invalid value fails."""
    with pytest.raises(ValueError):
        RealParam.with_value("invalid")


def test_create_int_param_with_value():
    """Test that an integer parameter can be created with a value."""
    param = IntParam.with_value(1)
    assert param.is_value_set()
    assert param.value == 1


def test_create_int_param_with_value_fails_with_invalid_value():
    """Test that creating an integer parameter with an invalid value fails."""
    with pytest.raises(ValueError):
        IntParam.with_value(1.2)


def test_param_bind_creates_assignment():
    """Test that binding a parameter creates an immutable assignment."""
    param = IntParam.with_lower_bound(0)
    assignment = param.bind(3)
    assert isinstance(assignment, ParamAssignment)
    assert assignment.value == 3
    assert assignment.param is param


def test_param_assignment_materialize_returns_bound_param():
    """Test that materialize returns the parameter definition."""
    assignment = RealParam.with_upper_bound(2.0).bind(1.5)
    bound_param = assignment.materialize()
    assert bound_param is assignment.param


def test_add_and_check_real_param_constraints(default_real_param):
    """Test that a real constraint can be added and checked."""
    default_real_param = default_real_param.add_constraint(
        EquationConstraint(
            default_real_param.variable,
            default_real_param.variable_expression * 3.14 < 20.0,
        )
    )
    default_real_param = default_real_param.add_constraint(
        EquationConstraint(
            default_real_param.variable, default_real_param.variable_expression >= 1.0
        )
    )
    _assert_all_satisfied(default_real_param, [2.0])
    _assert_none_satisfied(default_real_param, [0.5, 7.0])


def test_add_multiple_constraints_at_once():
    """Test that multiple constraints can be added in one call."""
    param = RealParam()
    constraints = [
        EquationConstraint(param.variable, param.variable_expression >= 1.0),
        EquationConstraint(param.variable, param.variable_expression <= 3.0),
    ]
    param = param.add_constraints(constraints)
    _assert_all_satisfied(param, [1.0, 2.0, 3.0])
    _assert_none_satisfied(param, [0.5, 3.5])


def test_add_multiple_constraints_validates_subclass_constraint_rules():
    """Test that add_constraints enforces subclass-specific constraint checks."""
    param = OrdinalParam([1, 2, 3])
    with pytest.raises(ValueError):
        param.add_constraints(
            [
                InSetConstraint(param.variable, {1, 2}),
                EquationConstraint(param.variable, param.variable_expression > 1),
            ]
        )


def test_add_and_check_int_param_constraints(default_int_param):
    """Test that an integer constraint can be added and checked."""
    default_int_param = default_int_param.add_constraint(
        EquationConstraint(
            default_int_param.variable,
            (default_int_param.variable_expression % 5).equals(0),
        )
    )
    default_int_param = default_int_param.add_constraint(
        EquationConstraint(
            default_int_param.variable, default_int_param.variable_expression > 10
        )
    )
    assert default_int_param.is_constraints_satisfied(15)
    with pytest.raises(ValueError):
        default_int_param = default_int_param.set_value(12)


def test_set_real_param_and_real_param_is_subset():
    """Test assignment values can be set independently from the same parameter."""
    param1 = RealParam()
    assignment_1 = param1.set_value(1.0)
    assignment_2 = param1.set_value(1.0)
    assert assignment_1.value == 1.0
    assert assignment_2.value == 1.0
    assert assignment_1.param is param1
    assert assignment_2.param is param1


def test_real_param_is_subset_of_real_param():
    """Test that a real parameter is a subset of another real parameter."""
    param1 = RealParam()
    param2 = RealParam()
    assert param1.is_subset(param2)
    assert param2.is_subset(param1)


def test_positive_real_param_and_real_param_is_subset():
    """Test subset relationship between positive real parameter and real parameter."""
    param1 = RealParam()
    param1 = param1.add_constraint(
        EquationConstraint(param1.variable, param1.variable_expression > 0)
    )
    param2 = RealParam()
    assert param1.is_subset(param2)
    assert not param2.is_subset(param1)


def test_interval_real_param_and_interval_real_param_is_subset():
    """Test subset relationship between interval real parameters."""
    param1 = RealParam()
    param1 = param1.add_constraint(
        EquationConstraint(param1.variable, param1.variable_expression >= 0)
    )
    param1 = param1.add_constraint(
        EquationConstraint(param1.variable, param1.variable_expression <= 3)
    )
    param2 = RealParam()
    param2 = param2.add_constraint(
        EquationConstraint(param2.variable, param2.variable_expression >= 0)
    )
    param2 = param2.add_constraint(
        EquationConstraint(param2.variable, param2.variable_expression <= 2)
    )
    assert not param1.is_subset(param2)
    assert param2.is_subset(param1)


def test_nat_param_zero_included():
    """Test that a natural number parameter with zero included can be set."""
    param = NatParam()
    assignment_0 = param.set_value(0)
    assert assignment_0.is_value_set()
    assignment_1 = param.set_value(1)
    assert assignment_1.is_value_set()
    with pytest.raises(ValueError):
        param.set_value(-1)


def test_nat_param_zero_excluded():
    """Test that a natural number parameter with zero excluded can be set."""
    param = NatParam(is_zero_included=False)
    assignment_1 = param.set_value(1)
    assert assignment_1.is_value_set()
    with pytest.raises(ValueError):
        param.set_value(0)
    with pytest.raises(ValueError):
        param.set_value(-1)


def test_nat_param_add_constraint_preserves_zero_excluded_state():
    """Test that adding constraints preserves zero-excluded NatParam semantics."""
    param = NatParam(is_zero_included=False)
    updated = param.add_lower_bound_constraint(2, is_inclusive=True)
    with pytest.raises(ValueError):
        updated.add_lower_bound_constraint(0, is_inclusive=True)


def test_ordinal_param_initialization():
    """Test that an ordinal parameter can be initialized."""
    param = OrdinalParam([5, 6, 7])
    assert isinstance(param, OrdinalParam)


def test_ordinal_param_initialization_fails_with_non_unique_values():
    """Test that an ordinal parameter initialization fails with non-unique values."""
    with pytest.raises(ValueError):
        OrdinalParam([1, 2, 1])


@pytest.fixture()
def ordinal_param_123() -> OrdinalParam:
    return OrdinalParam([1, 2, 3])


def test_set_ordinal_param_value(ordinal_param_123: OrdinalParam):
    """Test that an ordinal parameter value can be set."""
    assignment_1 = ordinal_param_123.set_value(1)
    assert assignment_1.is_value_set()
    assignment_3 = ordinal_param_123.set_value(3)
    assert assignment_3.is_value_set()


def test_set_ordinal_param_value_fails_with_invalid_value(
    ordinal_param_123: OrdinalParam,
):
    """Test that setting an ordinal parameter value fails with an invalid value."""
    with pytest.raises(ValueError):
        ordinal_param_123 = ordinal_param_123.set_value(4)


def test_add_and_check_ordinal_param_constraints(ordinal_param_123: OrdinalParam):
    """Test that ordinal parameter constraints can be added and checked."""
    ordinal_param_123 = ordinal_param_123.add_constraint(
        InSetConstraint(ordinal_param_123.variable, {1, 2})
    )
    _assert_all_satisfied(ordinal_param_123, [1, 2])
    _assert_none_satisfied(ordinal_param_123, [3])


def test_adding_invalid_constraint_to_ordinal_param_fails(
    ordinal_param_123: OrdinalParam,
):
    """Test that adding an invalid constraint to an ordinal parameter fails."""
    with pytest.raises(ValueError):
        ordinal_param_123 = ordinal_param_123.add_constraint(
            EquationConstraint(
                ordinal_param_123.variable, ordinal_param_123.variable_expression > 1
            )
        )


def test_categorical_param_initialization():
    """Test that a categorical parameter can be initialized."""
    param = CategoricalParam({"a", "b", "c"})
    assert isinstance(param, CategoricalParam)


@pytest.fixture()
def categorical_param_abc() -> CategoricalParam:
    return CategoricalParam({"a", "b", "c"})


def test_set_categorical_param_value(categorical_param_abc: CategoricalParam):
    """Test that a categorical parameter value can be set."""
    assignment_a = categorical_param_abc.set_value("a")
    assert assignment_a.is_value_set()
    assignment_c = categorical_param_abc.set_value("c")
    assert assignment_c.is_value_set()


def test_set_categorical_param_value_fails_with_invalid_value(
    categorical_param_abc: CategoricalParam,
):
    """Test that setting a categorical parameter value fails with an invalid value."""
    with pytest.raises(ValueError):
        categorical_param_abc = categorical_param_abc.set_value("d")


def test_add_and_check_categorical_param_constraints(
    categorical_param_abc: CategoricalParam,
):
    """Test that categorical parameter constraints can be added and checked."""
    categorical_param_abc = categorical_param_abc.add_constraint(
        InSetConstraint(categorical_param_abc.variable, {"a", "b"})
    )
    _assert_all_satisfied(categorical_param_abc, ["a", "b"])
    _assert_none_satisfied(categorical_param_abc, ["c"])


def test_adding_invalid_constraint_to_categorical_param_fails(
    categorical_param_abc: CategoricalParam,
):
    """Test that adding an invalid constraint to a categorical parameter fails."""
    with pytest.raises(ValueError):
        categorical_param_abc = categorical_param_abc.add_constraint(
            EquationConstraint(
                categorical_param_abc.variable,
                categorical_param_abc.variable_expression > "a",
            )
        )


def test_perm_param_initialization():
    """Test that a permutation parameter can be initialized."""
    param = PermParam(["n", "c", "h", "w"])
    assert isinstance(param, PermParam)


def test_perm_param_initialization_fails_with_non_unique_values():
    """Test that a permutation parameter initialization fails with non-unique values."""
    with pytest.raises(ValueError):
        PermParam(["n", "c", "h", "n"])


@pytest.fixture()
def perm_param_nchw() -> PermParam:
    return PermParam(["n", "c", "h", "w"])


def test_set_perm_param_value(perm_param_nchw: PermParam):
    """Test that a permutation parameter value can be set."""
    assignment = perm_param_nchw.set_value(["c", "n", "w", "h"])
    assert assignment.is_value_set()


def test_set_perm_param_value_fails_with_invalid_value(perm_param_nchw: PermParam):
    """Test that setting a permutation parameter value fails with an invalid value."""
    with pytest.raises(ValueError):
        perm_param_nchw = perm_param_nchw.set_value(["n", "c", "h", "n"])


def test_add_and_check_perm_param_constraints(perm_param_nchw: PermParam):
    """Test that permutation parameter constraints can be added and checked."""
    perm_param_nchw = perm_param_nchw.add_constraint(
        InSetConstraint(
            perm_param_nchw.variable, {("n", "c", "h", "w"), ("c", "n", "w", "h")}
        )
    )
    _assert_all_satisfied(perm_param_nchw, [["n", "c", "h", "w"], ["c", "n", "w", "h"]])
    _assert_none_satisfied(perm_param_nchw, [["n", "c", "w", "h"]])


def test_adding_invalid_constraint_to_perm_param_fails(perm_param_nchw: PermParam):
    """Test that adding an invalid constraint to a permutation parameter fails."""
    with pytest.raises(ValueError):
        perm_param_nchw = perm_param_nchw.add_constraint(
            EquationConstraint(
                perm_param_nchw.variable, perm_param_nchw.variable_expression > 1
            )
        )


def test_add_constraint_returns_new_param_without_mutating_original():
    """Test that adding constraints returns a new parameter definition."""
    param = OrdinalParam([1, 2, 3])
    updated = param.add_constraint(InSetConstraint(param.variable, {1, 2}))
    _assert_all_satisfied(updated, [1, 2])
    _assert_none_satisfied(updated, [3])
    _assert_all_satisfied(param, [1, 2, 3])


@pytest.mark.parametrize(
    "factory, ops, pass_values, fail_values",
    [
        # --- lower bound (mutating) ---
        pytest.param(
            partial(RealParam),
            [
                ("add_lower_bound_constraint", (1.0,), {"is_inclusive": True}),
            ],
            [1.0, 2.0],
            [0.5],
            id="real-lower-mutating-inclusive-1.0",
        ),
        pytest.param(
            partial(RealParam),
            [("add_lower_bound_constraint", (1.0,), {"is_inclusive": False})],
            [1.5, 2.0],
            [1.0, 0.5],
            id="real-lower-mutating-exclusive-1.0",
        ),
        # --- lower bound (constructor) ---
        pytest.param(
            partial(RealParam.with_lower_bound, 1.0, is_inclusive=True),
            [],
            [1.0, 2.0],
            [0.5],
            id="real-lower-constructor-inclusive-1.0",
        ),
        pytest.param(
            partial(RealParam.with_lower_bound, 1.0, is_inclusive=False),
            [],
            [1.5, 2.0],
            [1.0, 0.5],
            id="real-lower-constructor-exclusive-1.0",
        ),
        # --- upper bound (mutating) ---
        pytest.param(
            partial(RealParam),
            [
                ("add_upper_bound_constraint", (2.0,), {"is_inclusive": True}),
            ],
            [2.0, 1.0],
            [2.5],
            id="real-upper-mutating-inclusive-2.0",
        ),
        pytest.param(
            partial(RealParam),
            [("add_upper_bound_constraint", (2.0,), {"is_inclusive": False})],
            [1.0, 1.5],
            [2.0, 2.5],
            id="real-upper-mutating-exclusive-2.0",
        ),
        # --- upper bound (constructor) ---
        pytest.param(
            partial(RealParam.with_upper_bound, 2.0, is_inclusive=True),
            [],
            [2.0, 1.0],
            [2.5],
            id="real-upper-constructor-inclusive-2.0",
        ),
        pytest.param(
            partial(RealParam.with_upper_bound, 2.0, is_inclusive=False),
            [],
            [1.0, 1.5],
            [2.0, 2.5],
            id="real-upper-constructor-exclusive-2.0",
        ),
        # --- between (mutating) ---
        pytest.param(
            partial(RealParam),
            [
                ("add_lower_bound_constraint", (1.0,), {"is_inclusive": True}),
                ("add_upper_bound_constraint", (2.0,), {"is_inclusive": True}),
            ],
            [1.0, 1.5, 2.0],
            [0.5, 2.5],
            id="real-between-mutating-[1,2]",
        ),
        pytest.param(
            partial(RealParam),
            [
                ("add_lower_bound_constraint", (1.0,), {"is_inclusive": False}),
                ("add_upper_bound_constraint", (2.0,), {"is_inclusive": False}),
            ],
            [1.5],
            [1.0, 2.0, 0.5, 2.5],
            id="real-between-mutating-(1,2)",
        ),
        # --- between (constructor) ---
        pytest.param(
            partial(
                RealParam.between,
                1.0,
                2.0,
                is_lower_inclusive=True,
                is_upper_inclusive=True,
            ),
            [],
            [1.0, 1.5, 2.0],
            [0.5, 2.5],
            id="real-between-constructor-[1,2]",
        ),
        pytest.param(
            partial(
                RealParam.between,
                1.0,
                2.0,
                is_lower_inclusive=False,
                is_upper_inclusive=False,
            ),
            [],
            [1.5],
            [1.0, 2.0, 0.5, 2.5],
            id="real-between-constructor-(1,2)",
        ),
        # --- lower bound (mutating) ---
        pytest.param(
            partial(IntParam),
            [
                ("add_lower_bound_constraint", (1,), {"is_inclusive": True}),
            ],
            [1, 2],
            [0],
            id="int-lower-mutating-inclusive-1",
        ),
        pytest.param(
            partial(IntParam),
            [("add_lower_bound_constraint", (1,), {"is_inclusive": False})],
            [2, 5],
            [-1, 0],
            id="int-lower-mutating-exclusive-1",
        ),
        # --- lower bound (constructor) ---
        pytest.param(
            partial(IntParam.with_lower_bound, 1, is_inclusive=True),
            [],
            [1, 2],
            [0],
            id="int-lower-constructor-inclusive-1",
        ),
        pytest.param(
            partial(IntParam.with_lower_bound, 1, is_inclusive=False),
            [],
            [2, 5],
            [1, 0],
            id="int-lower-constructor-exclusive-1",
        ),
        # --- upper bound (mutating) ---
        pytest.param(
            partial(IntParam),
            [
                ("add_upper_bound_constraint", (2,), {"is_inclusive": True}),
            ],
            [2, 1],
            [3],
            id="int-upper-mutating-inclusive-2",
        ),
        pytest.param(
            partial(IntParam),
            [("add_upper_bound_constraint", (2,), {"is_inclusive": False})],
            [0, 1],
            [2, 3],
            id="int-upper-mutating-exclusive-2",
        ),
        # --- upper bound (constructor) ---
        pytest.param(
            partial(IntParam.with_upper_bound, 2, is_inclusive=True),
            [],
            [2, 1],
            [3],
            id="int-upper-constructor-inclusive-2",
        ),
        pytest.param(
            partial(IntParam.with_upper_bound, 2, is_inclusive=False),
            [],
            [1, 1],
            [2, 2],
            id="int-upper-constructor-exclusive-2",
        ),
        # --- between (mutating) ---
        pytest.param(
            partial(IntParam),
            [
                ("add_lower_bound_constraint", (1,), {"is_inclusive": True}),
                ("add_upper_bound_constraint", (2,), {"is_inclusive": True}),
            ],
            [1, 2],
            [-1, 3],
            id="int-between-mutating-[1,2]",
        ),
        pytest.param(
            partial(IntParam),
            [
                ("add_lower_bound_constraint", (1,), {"is_inclusive": False}),
                ("add_upper_bound_constraint", (3,), {"is_inclusive": False}),
            ],
            [2],
            [1, 3, 5, -5],
            id="int-between-mutating-(1,3)",
        ),
        # --- between (constructor) ---
        pytest.param(
            partial(
                IntParam.between, 1, 2, is_lower_inclusive=True, is_upper_inclusive=True
            ),
            [],
            [1, 2],
            [0, 3],
            id="int-between-constructor-[1,2]",
        ),
        pytest.param(
            partial(
                IntParam.between,
                1,
                3,
                is_lower_inclusive=False,
                is_upper_inclusive=False,
            ),
            [],
            [2],
            [1, 0, 3],
            id="int-between-constructor-(1,3)",
        ),
    ],
)
def test_numeric_params_adding_lower_and_upper_bounds(
    factory, ops, pass_values, fail_values
):
    """Test adding lower and upper bound constraints to int and real parameters."""
    param = factory()
    for name, args, kwargs in ops:
        param = getattr(param, name)(*args, **kwargs)
    _assert_all_satisfied(param, pass_values)
    _assert_none_satisfied(param, fail_values)


@pytest.mark.parametrize(
    "factory, ops",
    [
        pytest.param(
            partial(RealParam),
            [("add_lower_bound_constraint", ("invalid",))],
            id="real-lower-mutating-invalid",
        ),
        pytest.param(
            partial(RealParam),
            [("add_upper_bound_constraint", ("invalid",))],
            id="real-upper-mutating-invalid",
        ),
        pytest.param(
            partial(RealParam.with_upper_bound, "invalid"),
            [],
            id="real-upper-constructor-invalid",
        ),
        pytest.param(
            partial(RealParam.with_lower_bound, "invalid"),
            [],
            id="real-lower-constructor-invalid",
        ),
        pytest.param(
            partial(RealParam.between, 2.0, 1.0),
            [],
            id="real-between-constructor-invalid",
        ),
        pytest.param(
            partial(IntParam.between, 2, 1),
            [],
            id="int-between-constructor-invalid",
        ),
        pytest.param(
            partial(NatParam, is_zero_included=False),
            [("add_lower_bound_constraint", (0,))],
            id="nat-lower-mutating-invalid-zero-excluded",
        ),
        pytest.param(
            partial(NatParam, is_zero_included=False),
            [("add_lower_bound_constraint", (-1,))],
            id="nat-lower-mutating-invalid-negative-bound",
        ),
        pytest.param(
            partial(NatParam, is_zero_included=True),
            [("add_lower_bound_constraint", (-1,))],
            id="nat-lower-mutating-invalid-zero-included-exclusive",
        ),
        pytest.param(
            partial(NatParam, is_zero_included=False),
            [("add_upper_bound_constraint", (0,))],
            id="nat-upper-mutating-invalid-zero-included-exclusive",
        ),
        pytest.param(
            partial(NatParam, is_zero_included=False),
            [("add_upper_bound_constraint", (-1,))],
            id="nat-upper-mutating-invalid-zero-excluded",
        ),
        pytest.param(
            partial(NatParam, is_zero_included=True),
            [("add_upper_bound_constraint", (-1,))],
            id="nat-upper-mutating-invalid-negative-bound",
        ),
        pytest.param(
            partial(
                NatParam.between,
                -1,
                1,
                is_lower_inclusive=False,
                is_upper_inclusive=False,
            ),
            [],
            id="nat-between-constructor-invalid-negative-bound",
        ),
        pytest.param(
            partial(
                NatParam.between,
                3,
                2,
                is_lower_inclusive=False,
                is_upper_inclusive=False,
            ),
            [],
            id="nat-between-constructor-invalid-reversed-bound",
        ),
    ],
)
def test_numeric_params_adding_invalid_lower_and_upper_bounds_fails(factory, ops):
    with pytest.raises(ValueError):
        param = factory()
        for name, args, kwargs in ops:
            param = getattr(param, name)(*args, **kwargs)


def test_zero_included_in_nat_param_with_lower_bound_inclusive_at_zero():
    """Test that creating a NatParam with lower bound 0 inclusive works.

    This is specifically to test the constructor argument for including zero
    in the natural number constraints added before the lower bound constraint
    is added.
    """
    param = NatParam.with_lower_bound(0, is_inclusive=True)
    _assert_all_satisfied(param, [0, 1, 2, 100])


def test_real_param_serialization():
    """Test a real parameter can be serialized/deserialized via a dictionary."""
    param = RealParam()
    param = param.add_constraint(
        EquationConstraint(param.variable, param.variable_expression > 0)
    )
    param = param.add_constraint(
        EquationConstraint(param.variable, param.variable_expression < 10)
    )
    dictionary = param.serialize_to_dict()
    param2 = RealParam.deserialize_from_dict(dictionary)
    _assert_all_satisfied(param2, [1.0, 5.0, 9.0])
    _assert_none_satisfied(param2, [0.0, 10.0])


def test_int_param_serialization():
    """Test an int parameter can be serialized/deserialized via a dictionary."""
    param = IntParam()
    param = param.add_constraint(
        EquationConstraint(param.variable, param.variable_expression > 0)
    )
    param = param.add_constraint(
        EquationConstraint(param.variable, param.variable_expression < 10)
    )
    dictionary = param.serialize_to_dict()
    param2 = IntParam.deserialize_from_dict(dictionary)
    _assert_all_satisfied(param2, [1, 5, 9])
    _assert_none_satisfied(param2, [0, 10])


def test_ordinal_param_serialization(ordinal_param_123: OrdinalParam):
    """Test an ordinal parameter can be serialized/deserialized via a dictionary."""
    ordinal_param_123 = ordinal_param_123.add_constraint(
        InSetConstraint(ordinal_param_123.variable, {1, 2})
    )
    dictionary = ordinal_param_123.serialize_to_dict()
    param2 = OrdinalParam.deserialize_from_dict(dictionary)
    _assert_all_satisfied(param2, [1, 2])
    _assert_none_satisfied(param2, [3])


def test_categorical_param_serialization(categorical_param_abc: CategoricalParam):
    """Test a categorical parameter can be serialized/deserialized via a dictionary."""
    categorical_param_abc = categorical_param_abc.add_constraint(
        InSetConstraint(categorical_param_abc.variable, {"a", "b"})
    )
    dictionary = categorical_param_abc.serialize_to_dict()
    param2 = CategoricalParam.deserialize_from_dict(dictionary)
    _assert_all_satisfied(param2, ["a", "b"])
    _assert_none_satisfied(param2, ["c"])


def test_perm_param_serialization(perm_param_nchw: PermParam):
    """Test a permutation parameter can be serialized/deserialized via a dictionary."""
    perm_param_nchw = perm_param_nchw.add_constraint(
        InSetConstraint(
            perm_param_nchw.variable, {("n", "c", "h", "w"), ("c", "n", "w", "h")}
        )
    )
    dictionary = perm_param_nchw.serialize_to_dict()
    param2 = PermParam.deserialize_from_dict(dictionary)
    _assert_all_satisfied(param2, [["n", "c", "h", "w"], ["c", "n", "w", "h"]])
    _assert_none_satisfied(param2, [["n", "c", "w", "h"]])


def test_nat_param_serialization():
    """Test a nat parameter can be serialized/deserialized via a dictionary."""
    param = NatParam(is_zero_included=False)
    param = param.add_constraint(
        EquationConstraint(param.variable, param.variable_expression >= 1)
    )
    param = param.add_constraint(
        EquationConstraint(param.variable, param.variable_expression <= 10)
    )
    dictionary = param.serialize_to_dict()
    assert len(dictionary["__data__"]["constraints"]) == 3
    param2 = NatParam.deserialize_from_dict(dictionary)
    _assert_all_satisfied(param2, [1, 5, 10])
    _assert_none_satisfied(param2, [0, 11])
    dictionary = param2.serialize_to_dict()
    assert len(dictionary["__data__"]["constraints"]) == 3


# TODO: Check serialization structure errors and value errors for all types.


# TODO: Test the repr method.


def test_bound_int_param_between_inclusive_inclusive_constraints_satisfied():
    """Test ``between'' with inclusive bounds.

    Integer semantics: [3,5] => {3,4,5}
    """
    p = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=True)
    _assert_all_satisfied(p, [3, 4, 5])
    _assert_none_satisfied(p, [2, 6])


def test_bound_int_param_between_exclusive_exclusive_constraints_satisfied():
    """Test ``between'' with exclusive bounds.

    Integer semantics: (3,5) => only {4}
    """
    p = BoundIntParam.between(3, 5, is_lower_inclusive=False, is_upper_inclusive=False)
    _assert_all_satisfied(p, [4])
    _assert_none_satisfied(p, [3, 5, 2, 6])


def test_bound_int_param_between_exclusive_inclusive_constraints_satisfied():
    """Test ``between'' with exclusive lower bound and inclusive upper bound.

    Integer semantics: (3,5] => {4,5}
    """
    p = BoundIntParam.between(3, 5, is_lower_inclusive=False, is_upper_inclusive=True)
    _assert_all_satisfied(p, [4, 5])
    _assert_none_satisfied(p, [3, 2, 6])


def test_bound_int_param_between_inclusive_exclusive_constraints_satisfied():
    """Test ``between'' with inclusive lower bound and exclusive upper bound.

    Integer semantics: [3,5) => {3,4}
    """
    p = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=False)
    _assert_all_satisfied(p, [3, 4])
    _assert_none_satisfied(p, [5, 2, 6])


def test_bound_int_param_between_invalid_interval_raises_for_strict_equal_bounds():
    """Test that ``between'' raises for intervals with strict equal bounds."""
    with pytest.raises(ValueError):
        BoundIntParam.between(3, 3, is_lower_inclusive=False, is_upper_inclusive=False)


def test_bound_int_param_between_equal_bounds_inclusive_is_singleton():
    """Test that ``between'' with equal inclusive bounds is a singleton."""
    p = BoundIntParam.between(3, 3, is_lower_inclusive=True, is_upper_inclusive=True)
    _assert_all_satisfied(p, [3])
    _assert_none_satisfied(p, [2, 4])


def test_bound_int_param_between_invalid_order_raises_error():
    """Test that ``between'' raises for intervals with reversed bounds."""
    with pytest.raises(ValueError):
        BoundIntParam.between(5, 3)


def test_bound_int_param_with_lower_bound_inclusive():
    """Test ``with_lower_bound'' with inclusive bound."""
    p = BoundIntParam.with_lower_bound(3, is_inclusive=True)
    _assert_all_satisfied(p, [3, 4, 100])
    _assert_none_satisfied(p, [2, -1])


def test_bound_int_param_with_lower_bound_exclusive_integer_semantics():
    """Test ``with_lower_bound'' with exclusive bound.

    Integer semantics: x > 3 => ints {4,5,...}
    """
    p = BoundIntParam.with_lower_bound(3, is_inclusive=False)
    _assert_all_satisfied(p, [4, 5, 100])
    _assert_none_satisfied(p, [3, 2, -10])


def test_bound_int_param_with_upper_bound_inclusive():
    p = BoundIntParam.with_upper_bound(5, is_inclusive=True)
    _assert_all_satisfied(p, [5, 4, -100])
    _assert_none_satisfied(p, [6, 7])


def test_bound_int_param_with_upper_bound_exclusive_integer_semantics():
    """Test ``with_upper_bound'' with exclusive bound.

    Integer semantics: x < 5 => ints {...,3,4}
    """
    p = BoundIntParam.with_upper_bound(5, is_inclusive=False)
    _assert_all_satisfied(p, [4, 3, -100])
    _assert_none_satisfied(p, [5, 6])


def test_bound_int_param_exactly_satisfies_only_that_value():
    """Test ``exactly'' creates a singleton parameter."""
    p = BoundIntParam.exactly(7)
    _assert_all_satisfied(p, [7])
    _assert_none_satisfied(p, [6, 8, 0])


def test_bound_int_param_prefer_inclusive_flag_does_not_change_satisfiable_set():
    """Test that prefer_inclusive flag does not change satisfiable set."""
    p1 = BoundIntParam.between(
        3, 5, is_lower_inclusive=False, is_upper_inclusive=False, prefer_inclusive=True
    )
    p2 = BoundIntParam.between(
        3, 5, is_lower_inclusive=False, is_upper_inclusive=False, prefer_inclusive=False
    )
    for v in range(0, 10):
        assert p1.is_constraints_satisfied(v) == p2.is_constraints_satisfied(v)


def test_bound_int_param_set_value_accepts_int_only():
    """Test that BoundIntParam.set_value only accepts integer values."""
    p = BoundIntParam.with_lower_bound(0)
    assignment = p.set_value(1)
    assert assignment.value == 1
    with pytest.raises(ValueError):
        p.set_value(1.0)
    with pytest.raises(ValueError):
        p.set_value("1")


def test_bound_int_param_set_value_rejects_value_outside_constraints():
    """Test that ``set_value'' rejects values outside constraints."""
    p = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=True)
    with pytest.raises(ValueError):
        p.set_value(2)
    with pytest.raises(ValueError):
        p.set_value(6)
    assignment = p.set_value(4)
    assert assignment.value == 4


def test_bound_int_param_addition_of_singletons_is_singleton():
    """Test addition of two singleton BoundIntParams results in a singleton."""
    x = BoundIntParam.exactly(4)
    y = BoundIntParam.exactly(6)
    z = x + y
    _assert_all_satisfied(z, [10])
    _assert_none_satisfied(z, [9, 11])


def test_bound_int_param_addition_with_int_rhs():
    """Test addition of BoundIntParam with an integer on the right-hand side."""
    x = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=True)
    z = x + 2
    _assert_all_satisfied(z, [5, 6, 7])
    _assert_none_satisfied(z, [4, 8])


def test_bound_int_param_addition_with_int_lhs():
    """Test addition of BoundIntParam with an integer on the left-hand side."""
    x = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=True)
    z = 2 + x
    _assert_all_satisfied(z, [5, 6, 7])
    _assert_none_satisfied(z, [4, 8])


def test_bound_int_param_addition_bounds_propagate_semantics_from_strict_inputs():
    """Test addition bounds propagation with strict intervals.

    Integer semantics:
    x: (3,5) => {4}
    y: (5,10) => {6,7,8,9}
    x+y => {10,11,12,13}
    """
    x = BoundIntParam.between(3, 5, is_lower_inclusive=False, is_upper_inclusive=False)
    y = BoundIntParam.between(5, 10, is_lower_inclusive=False, is_upper_inclusive=False)
    z = x + y
    _assert_all_satisfied(z, [10, 11, 12, 13])
    _assert_none_satisfied(z, [9, 14])


def test_bound_int_param_addition_unbounded_lower_or_upper_propagates_unboundedness():
    """Test addition bounds propagation with unbounded inputs.

    Integer semantics:
    x >= 3, y unbounded => z >= 3 + (-inf) = -inf
    """
    x = BoundIntParam.with_lower_bound(3, is_inclusive=True)
    y = BoundIntParam()
    z = x + y
    _assert_all_satisfied(z, [-(10**6), 0, 10**6])


def test_bound_int_param_subtraction_of_singletons_is_singleton():
    """Test subtraction of two singleton BoundIntParams results in a singleton."""
    x = BoundIntParam.exactly(10)
    y = BoundIntParam.exactly(6)
    z = x - y
    _assert_all_satisfied(z, [4])
    _assert_none_satisfied(z, [3, 5])


def test_bound_int_param_subtraction_with_int_rhs():
    """Test subtraction of BoundIntParam with an integer on the right-hand side."""
    x = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=True)
    z = x - 2
    _assert_all_satisfied(z, [1, 2, 3])
    _assert_none_satisfied(z, [0, 4])


def test_bound_int_param_subtraction_with_int_lhs():
    """Test subtraction of BoundIntParam with an integer on the left-hand side."""
    x = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=True)
    z = 10 - x
    _assert_all_satisfied(z, [5, 6, 7])
    _assert_none_satisfied(z, [4, 8])


def test_bound_int_param_subtraction_bounds_propagate_semantics_from_strict_inputs():
    """Test subtraction bounds propagation with strict intervals.

    Integer semantics:
    x: (3,5) => {4}
    y: (5,10) => {6,7,8,9}
    x-y => {4-9 .. 4-6} = {-5,-4,-3,-2}
    """
    x = BoundIntParam.between(3, 5, is_lower_inclusive=False, is_upper_inclusive=False)
    y = BoundIntParam.between(5, 10, is_lower_inclusive=False, is_upper_inclusive=False)
    z = x - y
    _assert_all_satisfied(z, [-5, -4, -3, -2])
    _assert_none_satisfied(z, [-6, -1, 0])


def test_bound_int_param_negation_of_singleton():
    """Test negation of a singleton BoundIntParam results in a singleton."""
    x = BoundIntParam.exactly(4)
    z = -x
    _assert_all_satisfied(z, [-4])
    _assert_none_satisfied(z, [-3, -5])


def test_bound_int_param_negation_of_interval():
    """Test negation of an interval BoundIntParam."""
    x = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=True)
    z = -x
    _assert_all_satisfied(z, [-5, -4, -3])
    _assert_none_satisfied(z, [-6, -2])


def test_bound_int_param_negation_of_strict_interval_integer_semantics():
    """Test negation of a strict interval BoundIntParam."""
    x = BoundIntParam.between(3, 5, is_lower_inclusive=False, is_upper_inclusive=False)
    z = -x
    _assert_all_satisfied(z, [-4])
    _assert_none_satisfied(z, [-5, -3])


def test_bound_int_param_prefer_inclusive_changes_str_not_membership_for_addition():
    """Test that prefer_inclusive changes string but not membership for addition."""
    x_incl = BoundIntParam.between(
        3, 5, is_lower_inclusive=False, is_upper_inclusive=False, prefer_inclusive=True
    )
    y_incl = BoundIntParam.between(
        5, 10, is_lower_inclusive=False, is_upper_inclusive=False, prefer_inclusive=True
    )

    x_excl = BoundIntParam.between(
        3, 5, is_lower_inclusive=False, is_upper_inclusive=False, prefer_inclusive=False
    )
    y_excl = BoundIntParam.between(
        5,
        10,
        is_lower_inclusive=False,
        is_upper_inclusive=False,
        prefer_inclusive=False,
    )

    z_incl = x_incl + y_incl
    z_excl = x_excl + y_excl

    for v in range(-20, 40):
        assert z_incl.is_constraints_satisfied(v) == z_excl.is_constraints_satisfied(v)
    assert str(z_incl) != str(z_excl)


def test_bound_int_param_addition_accepts_int_param():
    """Test addition of BoundIntParam with IntParam on the right-hand side."""
    x = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=True)
    y = IntParam.between(5, 10, is_lower_inclusive=True, is_upper_inclusive=True)
    z = x + y
    _assert_all_satisfied(z, [8, 9, 10, 11, 12, 13, 14, 15])
    _assert_none_satisfied(z, [7, 16])


def test_bound_int_param_subtraction_accepts_int_param():
    """Test subtraction of BoundIntParam with IntParam on the right-hand side."""
    x = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=True)
    y = IntParam.between(5, 10, is_lower_inclusive=True, is_upper_inclusive=True)
    z = x - y
    _assert_all_satisfied(z, [-7, -3, 0])
    _assert_none_satisfied(z, [-8, 1])


def test_bound_int_param_rsub_accepts_int_param_on_left():
    """Test subtraction of BoundIntParam with IntParam on the left-hand side."""
    x = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=True)
    y = IntParam.between(5, 10, is_lower_inclusive=True, is_upper_inclusive=True)
    z = y - x
    _assert_all_satisfied(z, [0, 7])
    _assert_none_satisfied(z, [-1, 8])


def test_bound_int_param_addition_with_unsupported_type_raises_error():
    """Test that addition of BoundIntParam with unsupported type raises TypeError."""
    x = BoundIntParam.between(0, 1)
    with pytest.raises(TypeError):
        _ = x + "nope"


@pytest.mark.parametrize(
    "lower,upper,lin,u_in",
    [
        (0, 0, True, True),
        (0, 1, True, True),
        (0, 1, False, True),
        (0, 1, True, False),
        (0, 2, False, False),
        (-3, 3, True, True),
        (-3, 3, False, False),
    ],
)
def test_bound_int_param_addition_matches_brute_force(lower, upper, lin, u_in):
    """Test that addition matches brute-force set addition."""
    x = BoundIntParam.between(
        lower, upper, is_lower_inclusive=lin, is_upper_inclusive=u_in
    )
    y = BoundIntParam.between(
        lower, upper, is_lower_inclusive=lin, is_upper_inclusive=u_in
    )
    z = x + y
    allowed_x = [
        v for v in range(lower - 2, upper + 3) if x.is_constraints_satisfied(v)
    ]
    allowed_y = [
        v for v in range(lower - 2, upper + 3) if y.is_constraints_satisfied(v)
    ]
    allowed_z = {a + b for a in allowed_x for b in allowed_y}
    for v in range(2 * (lower - 2), 2 * (upper + 2) + 1):
        assert z.is_constraints_satisfied(v) == (v in allowed_z)


@pytest.mark.parametrize(
    "lower,upper,lin,u_in",
    [
        (0, 0, True, True),
        (0, 1, True, True),
        (0, 2, False, False),
        (-2, 2, True, True),
        (-2, 2, False, False),
    ],
)
def test_bound_int_param_subtraction_matches_brute_force(lower, upper, lin, u_in):
    """Test that subtraction matches brute-force set subtraction."""
    x = BoundIntParam.between(
        lower, upper, is_lower_inclusive=lin, is_upper_inclusive=u_in
    )
    y = BoundIntParam.between(
        lower, upper, is_lower_inclusive=lin, is_upper_inclusive=u_in
    )
    z = x - y
    allowed_x = [
        v for v in range(lower - 2, upper + 3) if x.is_constraints_satisfied(v)
    ]
    allowed_y = [
        v for v in range(lower - 2, upper + 3) if y.is_constraints_satisfied(v)
    ]
    allowed_z = {a - b for a in allowed_x for b in allowed_y}
    for v in range((lower - 2) - (upper + 2), (upper + 2) - (lower - 2) + 1):
        assert z.is_constraints_satisfied(v) == (v in allowed_z)


def test_bound_int_param_serialization():
    """Test bound int parameter can be serialized/deserialized via a dictionary."""
    p = BoundIntParam.between(3, 5, is_lower_inclusive=True, is_upper_inclusive=False)
    dictionary = p.serialize_to_dict()
    p2 = BoundIntParam.deserialize_from_dict(dictionary)
    _assert_all_satisfied(p2, [3, 4])
    _assert_none_satisfied(p2, [5])


def test_bound_nat_param_serialization():
    """Test bound nat parameter can be serialized/deserialized via a dictionary."""
    p = BoundNatParam.with_lower_bound(2, is_inclusive=True)
    dictionary = p.serialize_to_dict()
    assert len(dictionary["__data__"]["constraints"]) == 2
    p2 = BoundNatParam.deserialize_from_dict(dictionary)
    _assert_all_satisfied(p2, [2, 5, 100])
    _assert_none_satisfied(p2, [0, 1])
    dictionary = p2.serialize_to_dict()
    assert len(dictionary["__data__"]["constraints"]) == 2


# TODO: Update tests that check exceptions to match exception messages.



"""String utilities."""

__all__ = ["format_comma_separated_list"]

from collections.abc import Iterable
from typing import Callable, TypeVar

_T = TypeVar("_T")


def format_comma_separated_list(
    items: Iterable[_T], str_func: Callable[[_T], str] = repr, add_space: bool = True
) -> str:
    """Return a comma-separated list of items.

    Args:
        items: An iterable of items.
        str_func: A function to convert each item to a string.
        add_space: Whether to add a space after each comma.

    Returns:
        A comma-separated list of items.

    """
    join_char = ", " if add_space else ","
    return join_char.join(
        str_func(item) if not isinstance(item, str) else item for item in items
    )



"""Tests the logging utility."""

import logging
from pathlib import Path

import pytest
from fhy_core.logger import (
    add_file_handler,
    configure_logging,
    get_logger,
    install_null_handler,
    reset,
)


@pytest.fixture(autouse=True)
def isolate_logging_state():
    root_logger = logging.getLogger()
    root_logger_handlers = list(root_logger.handlers)
    root_logger_level = root_logger.level

    touched_names = set()

    def touch(name: str):
        touched_names.add(name)
        return logging.getLogger(name)

    yield touch

    root_logger.handlers[:] = root_logger_handlers
    root_logger.setLevel(root_logger_level)

    for name in touched_names:
        logger = logging.getLogger(name)
        logger.handlers[:] = []
        logger.setLevel(logging.NOTSET)
        logger.propagate = True

    reset()


def _get_non_file_stream_handlers(logger: logging.Logger):
    return [
        h
        for h in logger.handlers
        if isinstance(h, logging.StreamHandler)
        and not isinstance(h, logging.FileHandler)
    ]


def _get_file_handlers(logger: logging.Logger):
    return [h for h in logger.handlers if isinstance(h, logging.FileHandler)]


def test_get_logger_returns_named_logger():
    """Test `get_logger` returns a logger with the correct name."""
    name = "some.pkg.mod"

    logger = get_logger(name)

    assert isinstance(logger, logging.Logger)
    assert logger.name == name


def test_get_logger_does_not_configure_handlers_or_level():
    """Test `get_logger` does not modify root logger configuration."""
    root_logger = logging.getLogger()
    before_handlers = list(root_logger.handlers)
    before_level = root_logger.level

    _ = get_logger("x.y.z")

    assert list(root_logger.handlers) == before_handlers
    assert root_logger.level == before_level


def test_install_null_handler_adds_only_if_no_handlers(isolate_logging_state):
    """Test `install_null_handler` adds a NullHandler only if no handlers exist."""
    namespace = "my_namespace"
    logger = isolate_logging_state(namespace)
    assert logger.handlers == []

    install_null_handler(namespace)

    assert len(logger.handlers) == 1
    assert isinstance(logger.handlers[0], logging.NullHandler)


def test_install_null_handler_does_not_add_if_handlers_exist(isolate_logging_state):
    """Test `install_null_handler` does not add a NullHandler if handlers exist."""
    namespace = "my_namespace"
    logger = isolate_logging_state(namespace)

    sentinel = logging.StreamHandler()
    logger.addHandler(sentinel)

    install_null_handler(namespace)

    assert logger.handlers == [sentinel]


def test_install_null_handler_is_idempotent(isolate_logging_state):
    """Test `install_null_handler` is idempotent."""
    namespace = "my_namespace"
    logger = isolate_logging_state(namespace)

    install_null_handler(namespace)
    install_null_handler(namespace)

    assert len(logger.handlers) == 1
    assert isinstance(logger.handlers[0], logging.NullHandler)


def test_configure_logging_adds_single_console_handler_and_sets_levels(
    isolate_logging_state,
):
    """Test `configure_logging` adds a console handler with correct level."""
    namespace = "mycompiler"
    logger = isolate_logging_state(namespace)

    out = configure_logging(namespace, console_level=logging.WARNING)

    assert out is logger
    stream_handlers = _get_non_file_stream_handlers(logger)
    assert len(stream_handlers) == 1
    assert stream_handlers[0].level == logging.WARNING
    assert _get_file_handlers(logger) == []


def test_configure_logging_is_idempotent_for_console_handler(isolate_logging_state):
    """Test `configure_logging` does not duplicate handlers on multiple calls."""
    namespace = "mycompiler"
    logger = isolate_logging_state(namespace)

    _ = configure_logging(namespace, console_level=logging.INFO)
    _ = configure_logging(namespace, console_level=logging.ERROR)

    stream_handlers = _get_non_file_stream_handlers(logger)
    assert len(stream_handlers) == 1
    assert stream_handlers[0].level == logging.ERROR


def test_configure_logging_adds_file_handler_and_is_idempotent_by_resolved_path(
    isolate_logging_state, tmp_path
):
    """Test `configure_logging` adds file handler and does not duplicate."""
    namespace = "mycompiler"
    logger = isolate_logging_state(namespace)

    log_path = tmp_path / "build.log"
    format1 = logging.Formatter("A:%(message)s")
    format2 = logging.Formatter("B:%(message)s")

    configure_logging(
        namespace,
        file_path=log_path,
        file_level=logging.DEBUG,
        formatter=format1,
    )
    configure_logging(
        namespace,
        file_path=str(log_path),
        file_level=logging.ERROR,
        formatter=format2,
    )

    file_handlers = _get_file_handlers(logger)
    assert len(file_handlers) == 1
    assert file_handlers[0].level == logging.ERROR
    assert file_handlers[0].formatter is format2
    assert Path(file_handlers[0].baseFilename).resolve() == log_path.resolve()


def test_configure_logging_does_not_duplicate_if_existing_streamhandler_present(
    isolate_logging_state,
):
    """Test `configure_logging` updates existing StreamHandler."""
    namespace = "mycompiler"
    logger = isolate_logging_state(namespace)

    existing = logging.StreamHandler()
    existing.setLevel(logging.CRITICAL)
    logger.addHandler(existing)

    configure_logging(namespace=namespace, console_level=logging.INFO)

    stream_handlers = _get_non_file_stream_handlers(logger)
    assert len(stream_handlers) == 1
    assert stream_handlers[0] is existing
    assert stream_handlers[0].level == logging.INFO


def test_add_file_handler_adds_and_is_idempotent(isolate_logging_state, tmp_path):
    """Test `add_file_handler` adds FileHandler and is idempotent by absolute path."""
    namespace = "mycompiler"
    logger = isolate_logging_state(namespace)

    log_path = tmp_path / "x.log"
    add_file_handler(logger, log_path, level=logging.WARNING)

    file_handlers = _get_file_handlers(logger)
    assert len(file_handlers) == 1
    assert file_handlers[0].level == logging.WARNING
    assert Path(file_handlers[0].baseFilename).resolve() == log_path.resolve()

    add_file_handler(logger, str(log_path), level=logging.ERROR)

    file_handlers2 = _get_file_handlers(logger)
    assert len(file_handlers2) == 1
    assert file_handlers2[0].level == logging.ERROR


def test_add_file_handler_uses_explicit_formatter_when_provided(
    isolate_logging_state, tmp_path
):
    """Test that `add_file_handler` uses the provided formatter."""
    logger = isolate_logging_state("mycompiler")

    formatter = logging.Formatter("X:%(message)s")
    log_path = tmp_path / "x.log"

    add_file_handler(logger, log_path, formatter=formatter)
    file_handler = _get_file_handlers(logger)[0]
    assert file_handler.formatter is formatter


def test_add_file_handler_infers_formatter_from_existing_handler(
    isolate_logging_state, tmp_path
):
    """Test `add_file_handler` uses existing handler's formatter if none provided."""
    logger = isolate_logging_state("mycompiler")

    formatter = logging.Formatter("INFER:%(message)s")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    log_path = tmp_path / "x.log"
    add_file_handler(logger, log_path, formatter=None)

    file_handler = _get_file_handlers(logger)[0]
    assert file_handler.formatter is formatter


def test_configure_logging_propagate_flag(isolate_logging_state):
    """Test `configure_logging` sets the propagate flag correctly."""
    namespace = "mycompiler"
    logger = isolate_logging_state(namespace)

    configure_logging(namespace, propagate=True)
    assert logger.propagate is True

    configure_logging(namespace, propagate=False)
    assert logger.propagate is False


def test_stream_handler_detection_does_not_count_filehandler_as_console(
    isolate_logging_state, tmp_path
):
    """Test `configure_logging` detects existing handler when a handler exists."""
    namespace = "mycompiler"
    logger = isolate_logging_state(namespace)

    log_path = tmp_path / "only_file.log"
    file_handler = logging.FileHandler(str(log_path), encoding="utf-8")
    logger.addHandler(file_handler)

    configure_logging(namespace=namespace, console_level=logging.INFO)

    assert len(_get_file_handlers(logger)) == 1
    assert len(_get_non_file_stream_handlers(logger)) == 1
